                           (* MTSs *)
(** MTSs in OCaml: testing a process to merge inconsistent MTSs over the *same alphabet* using a heuristics. *)


(** Types *)

type trieval = True | Maybe | False;;
type label = trieval * string;;

(** A transitions_matrix m represents, for a given letter all the outgoing transition:
m.(s).(t) can be True, Maybe, or False, depending if there is a true, maybe, or no
transition at all from s to t on the letter for which m is the transitions_matrix.     *)
type transitions_matrix = trieval array array;;
type value_matrix = float array array;;

type mts = {
  alphabet:string array; (** The set of actions as it will be displayed. Do not use
			 '?' or ',' to avoid confusion (won't pass the check).      *)
  states: int; (** The states are numbered from 0 to mts.states-1 *)
  transitions: transitions_matrix array (** An array reprsenting the transitions for each action. The order
					is what makes the correspondance between the action as it will be
					displayed and the transitions associated with that letter.          *)
};;

type parseResult = {
  name : string ; (** The name of the MTS; useful for keeping track of printed MTSs. *)
  nbstates : int; (** The states are numbered from 0 to mts.states-1 *)
  trans : (int * ((((trieval * string) list) * int) list)) list (** The data of the transition, in a raw form from the parsed
								    transitions list. It is of the form:
				    (originState, ((True or Maybe,label of the transition), destinationState)::queue)::queue. *)
};;

(** Mu calculus formulas. *)
type muCalc =
  (* Constants *)
  | Empty
  | MuTrue
  | MuFalse
  (* Variables *)
  | Var of string
  (* Next operators *)
  | Enext of string * muCalc
  | Anext of string * muCalc
  (* Logical connectors *)
  | Not of muCalc
  | Or of muCalc * muCalc
  | And of muCalc * muCalc
  (* Fixpoint operators *)
  | Mu of string * muCalc
  | Nu of string * muCalc
;;

(* Various tools part I *)

(** max_list l returns the maximal value of a l.
Works for any type having a 'compare' function.  *)
let rec max_list = function
  |[] -> failwith "[max_list] Empty list"
  |[e] -> e
  |t::q -> max t (max_list q)
;;

(** int_exp a n returns a^n, a and n being integers. *)
let rec int_exp base expo =
  if expo = 0
  then
    1
  else
    let prev = int_exp base (expo / 2)
    in
      if expo mod 2 = 0
      then
	prev * prev
      else
	prev * prev * base
;;

(** float_round p x returns the approximation
of x having at most p digits after the point. *)  
let float_round (precision:int) (the_num:float) =
  let power = float_of_int (int_exp 10 precision)
  in
    let moved_point = int_of_float (the_num *. (power *. 10.))
    in
      let question_digit = moved_point mod 10
      in
	if question_digit < 5
	then
	  (float_of_int (moved_point / 10)) /. power
	else
	  (float_of_int ((moved_point / 10) + 1)) /. power
;;

(** insert i l returns the list l fi i did not
belong to it beore, otherwise, returns l.      *)
let rec insert i = function
  | [] -> [i]
  | t::q when t = i -> t::q
  | t::q -> t::(insert i q)
;;

(** initMts a n returns a fresh MTS (type mts) having
n states and maybe transitions on all letters of a
between every pair of states (including self loops).
This MTS is an abstraction of any other. *)
let initMts alpha nb_states =
  let fresh = {
    alphabet = Array.copy alpha;
    states = nb_states;
    transitions = Array.init (Array.length alpha) (fun x -> Array.init nb_states (fun x -> Array.make nb_states Maybe))
  }
  in
    fresh
;;

(** revertMts m returns a fresh MTS being the copy of m
but having all its arrows reverted. The states are
unchanged, therefore the initial state is still the same. *)
let revertMts (orig:mts) =
  let res = initMts orig.alphabet orig.states
  and alpha = Array.length orig.alphabet
  in
    for a = 0 to alpha-1 do
      for i = 0 to res.states-1 do
	for j = 0 to res.states-1 do
	  res.transitions.(a).(i).(j) <- orig.transitions.(a).(j).(i)
	done
      done
    done;
    res
;;

(** copyMts m returns a fresh (deep) copy of m. *)
let copyMts (orig:mts) =
  let res = initMts orig.alphabet orig.states
  and alpha = Array.length orig.alphabet
  in
    for a = 0 to alpha-1 do
      for i = 0 to res.states-1 do
	for j = 0 to res.states-1 do
	  res.transitions.(a).(i).(j) <- orig.transitions.(a).(i).(j)
	done
      done
    done;
    res
;;

(* let extendAlphabet (newVal:trieval) (old:mts) (newAlphabet:(string Array)) = *)
let extendAlphabet newVal old newAlphabet =
  let oldAlpha = Array.length old.alphabet
  and newAlpha = Array.length newAlphabet
  in
  let correspIndex oldArray newArray oldIndex =
    let len = Array.length newArray
    and curr = ref 0
    in
      while (!curr < len) && (newArray.(!curr) <> oldArray.(oldIndex)) do
	curr := !curr + 1
      done;
      if (!curr = len)
      then -1
      else !curr
  in
  let oldI = ref 0
  in
    while (!oldI < oldAlpha) && ((correspIndex old.alphabet newAlphabet !oldI) >= 0) do
      oldI := !oldI + 1
    done;
    if (!oldI < oldAlpha) then failwith "[extendAlphabet] First alphabet not included in the extension.";
    let initNew a =
      let oldIndex = correspIndex newAlphabet old.alphabet a
      in
	if (oldIndex < 0) then
	  Array.init old.states (fun i -> Array.make old.states newVal)
	else
	  Array.init old.states (fun i -> Array.init old.states (fun j -> old.transitions.(oldIndex).(i).(j)))
    in
    let res = {
      alphabet = Array.copy newAlphabet;
      states = old.states;
      transitions = Array.init newAlpha initNew
    }
    in
      res
;;

let arrayUnion alph1 alph2 =
  let rec union acc = function
    | [] -> acc
    | t::q -> union (insert t acc) q
  in
  let l1 = Array.to_list alph1
  and l2 = Array.to_list alph2
  in
  let lu = union l1 l2
  in
    Array.of_list lu
;;


(** get_outoing c m s returns the list of transitions in m
steming from s and whose value complies to condition c. The
output is a tuple composed of the index of the action in m, the
target, and the value of the transition (True, Maybe, or False). *)
let get_outgoing (condition:trieval -> bool) (m:mts) (source:int) =
  let alpha = Array.length m.alphabet
  in
  let rec get_out_rec target letter =
    match target,letter with
      |(-1),_ -> []
      |targ,(-1) -> get_out_rec (targ-1) (alpha-1)
      |targ,lett ->
	 let prev_result = get_out_rec targ (lett-1)
	 and the_trans = m.transitions.(lett).(source).(targ)
	 in
	   if (condition the_trans)
	   then
	     (lett,targ,the_trans)::prev_result
	   else
	     prev_result
  in
    get_out_rec (m.states-1) (alpha-1)
;;

(* Some useful conditions to be used with the above function *)
(** is_true, is_maybe, is_possible are conditions whose
result is straightforward according to their names.     *)
let is_true = function
  | True -> true
  | _ -> false
and is_maybe = function
  | Maybe -> true
  | _ -> false
and is_possible = function
  | False -> false
  | _ -> true
;;

(** getLabel mpm s returns the label attached to a state of a pseudomerge computed with keepTrack set to true. *)
let getLabel (merge:mts) (state:int) =
  let alpha = Array.length merge.alphabet
  in
  let labelsstart = alpha - merge.states
  and out = ref []
  in
    for a = labelsstart to alpha - 1 do
      if is_true merge.transitions.(a).(state).(state)
      then
	out := a::(!out)
    done;
    match (!out) with
      | [] -> failwith "[getLabel] This state has no label (is the merge really marked ?)."
      | a::[] -> merge.alphabet.(a)
      | _ -> failwith "[getLabel] This state has multiple labels."
;;

(** From a label l = __i_j__ parseLabel l returns the values i and j. *)
let parseLabel l =
  let i_j = String.sub l 2 ((String.length l) - 4)
  in
  let underPos = String.index i_j '_'
  in
  let iStr = String.sub i_j 0 underPos
  and jStr = String.sub i_j (underPos + 1) ((String.length i_j) - (underPos+1))
  in
  let i = ref 0
  and j = ref 0
  in
    (try
       i := int_of_string iStr
     with Failure "int_of_string" ->
       i := -1
    );
    (try
       j := int_of_string jStr
     with Failure "int_of_string" ->
       j := -1
    );
    (*print_string ("\t" ^ l ^ " -> (" ^ iStr ^ "," ^ jStr ^") -> ");
      printIntPair (!i,!j);*)
    !i,!j
;;

(** correspStates mpm s returns the pairof states state s in mpm came from during a merge.
An unknown state is denoted by -1. *)
let correspStates (merge:mts) (state:int) =
  parseLabel (getLabel merge state)
;;

(** isMStar mpm s returns true iff the state s is a disagreement state m,*. *)
let isMStar (merge:mts) (state:int) =
  let i,j = correspStates merge state
  in
    j < 0
;;

(** isStarN mpm s returns true iff the state s is a disagreement state *,n. *)
let isStarN (merge:mts) (state:int) =
  let i,j = correspStates merge state
  in
    i < 0
;;

(** isStar mpm s returns true iff the state s is a disagreement state. *)
let isStar (merge:mts) (state:int) =
  (isMStar merge state) || (isStarN merge state)
;;


(** loosestConsistency m n returns the list of pairs (i,j) for 0 <= i < m and 0 <= j < n. *)
let loosestConsistency (statesM:int) (statesN:int) =
  let rec routine accu m n=
    match m,n with
      | 0,0 -> (0,0)::accu
      | m,0 -> routine ((m,0)::accu) (m-1) (statesN - 1)
      | m,n -> routine ((m,n)::accu) m (n-1)
  in
    routine [] (statesM - 1) (statesN - 1)
;;

(* Pretty (or not) printing *)
(* These functions do not check the validity of the input *)

(** printIntArray a displays an arrays of integers separated by '|' *)
let printIntArray a =
  let n = Array.length a
  in
    print_string "\n|";
    for k = 0 to n-1 do
      print_string (" " ^ (string_of_int a.(k)) ^ " |")
    done
;;

(** build_hrule l n returns a horizontal rule to be used when
printing a table having n columns whose content is of length l. *)
let build_hrule cell_length cells_in_row =
  let refHrule = ref "\t"
  in
    for i = 1 to cells_in_row do
      refHrule := !refHrule ^ "+";
      for j = 1 to cell_length do
	refHrule := !refHrule ^ "-"
      done
    done;
    !refHrule ^ "+"
;;

(** split_space s f returns the number of blank spaces
to be placed before and after f so that it is (more
or less) centered in a space of total length s.        *)
let split_space cell_space fill = 
  let fill_space = String.length(fill)
  in
    let free_space = cell_space - fill_space
    in
      if free_space < 0
      then
	failwith "[split_space] Not enough space in the cell for all the text"
      else
	let bef = free_space/2
	in
	  bef,free_space-bef
;;

(** padding n returns a string formed of n blank spaces *)
let padding space = 
  let pad = ref ""
  in
    for k = 1 to space do
      pad := !pad ^ " "
    done;
    !pad
;;

(** transition_to_string m s t returns a string being a coma separated
list of maybe and true transitions from s to t in m. Maybe transitions
have a question mark appended to tha name of the action.               *)
let transition_to_string_gen (normal:bool) (m:mts) state_from state_to =
  let res = ref ""
  and alpha = 
    if normal then
      Array.length m.alphabet
    else
      begin
	let curr = ref 0
	and maxalpha = Array.length m.alphabet
	in
	  while ((!curr < maxalpha) && (m.alphabet.(!curr) <> "__0_0__")) do
	    curr :=  !curr + 1
	  done;
	  if (!curr = maxalpha)
	  then
	    failwith "[transition_to_string_gen] Marked action '__0_0__' was not found."
	  else
	    !curr
      end
  and first_letter = ref true
  in
    for a = 0 to alpha-1 do
      match m.transitions.(a).(state_from).(state_to) with
	| True ->
	    if !first_letter
	    then
	      begin
		res := !res ^ m.alphabet.(a);
		first_letter := false
	      end
	    else
		res := !res ^ "," ^ m.alphabet.(a)
	| Maybe ->
	    if !first_letter
	    then
	      begin
		res := !res ^ m.alphabet.(a) ^ "?";
		first_letter := false
	      end
	    else
	      res := !res ^ "," ^ m.alphabet.(a) ^ "?";
	| False -> ()
    done;
    !res
;;

let transition_to_string = transition_to_string_gen true
and unmarked_transition_to_string = transition_to_string_gen false
;;

(** string_of_trieval v returns a one character long string
representing the value True ("T"), Maybe ("?"), or False ("F"). *)
let string_of_trieval = function
  | True  -> "T"
  | Maybe -> "?"
  | False -> "F"
;;

let padded_of_trieval = function
  | True  -> "True "
  | False -> "False"
  | Maybe -> "Maybe"
;;

let extractAlphaString (lend:string) (rend:string) (sep:string) (m:mts) (isMerge:bool) =
  let res = ref ""
  and alpha = Array.length m.alphabet
  and first = ref true
  in
    for a = 0 to (alpha - 1) do
      if ((isMerge && (a mod 3 = 0)) || (not isMerge))
      then
	begin
	  if !first then
	    first := false
	  else
	    res := !res ^ sep
	  ;
	  res := !res ^ m.alphabet.(a)
	end
    done;
    lend ^ !res ^ rend
;;

(** printMts m outputs a representation of m as a table. Each row
represents a source state and each column a target state. Each cell
bears the representation of transitions between the source and the target
state as a coma separated list of actions, with '?'for maybe transitions. *)
let printMts (m:mts) =
  let refCell_length = ref 0
  and alpha = Array.length(m.alphabet)
  in
    for a = 0 to alpha-1 do
      (* The maximum length of a cell is reached when between these   *)
      (* states there are maybe transitions on every letter: we add 1 *)
      (* space for the '?' and one for the ',' separating the letters *)
      refCell_length := !refCell_length + String.length(m.alphabet.(a)) + 2
    done;
    let cell_length = max (!refCell_length - 1) (String.length(string_of_int(m.states)))
    in
      let hrule = build_hrule cell_length (m.states + 1)
      in
	print_string("\n"^hrule^"\n\t|" ^ (padding cell_length) ^ "|");
	for stateh = 0 to m.states-1 do
	  let to_print = string_of_int stateh
	  in
	    let bef,aft = split_space cell_length to_print
	    in
	      print_string((padding bef) ^ to_print ^ (padding aft) ^ "|")
	done;
	print_string("\n" ^ hrule ^ "\n");
	for statev = 0 to m.states-1 do
	  let to_print = string_of_int statev
	  in
	    let bef,aft = split_space cell_length to_print
	    in
	      print_string("\t|" ^ (padding bef) ^ to_print ^ (padding aft) ^ "|");
	      for stateh = 0 to m.states-1 do
		let cell_to_print = transition_to_string m statev stateh
		in
		  let cell_bef,cell_aft = split_space cell_length cell_to_print
		  in
		    print_string((padding cell_bef) ^ cell_to_print ^ (padding cell_aft) ^ "|")
	      done;
	      print_string("\n" ^ hrule ^ "\n")
	done
;;

(** printMtsLatex s n m outputs the LaTeX code for the MTS m. The code is a
tabular having the name n of te MTS on the left and a Vaucanson automaton
on the left. The states are aligned and separated by a distance of s. All
transitions are arcs (\ArcL[.5]{source}{target}), and self loops are oriented
north (\LoopN[.5]{state}). States are names Statei, and are labeled by i, i
being the number of the state in m. The positionning of the states and the
edges may (and surely will) require some work to have a readable MTS. The
name n can include LaTeX commands, but beware that the '\' need to be escaped. *)
let printMtsLatex (stateSep:float) (nom:string) (m:mts) =
  print_string ("\n\\begin{tabular}{cl}\n" ^
		  nom ^
		  ": & \\TinyPicture\\VCDraw{%\n\\begin{VCPicture}{(0,-.5)(" ^
		  (string_of_float ((float_of_int (m.states -1)) *. stateSep)) ^
		  ",0)}\n");
  (* States. Just in order to have placeholders; should be modified for a pretty printing. *)
  for i = 0 to m.states - 1 do
    print_string ("\\StateVar[" ^
		    (string_of_int i) ^
		    "]{(" ^
		    (string_of_float ((float_of_int i) *. stateSep)) ^
		    ",0)}{State" ^
		    (string_of_int i) ^
		    "}\n")
  done;
  (* Transitions. The shape might need some work. *)
  let alpha =  Array.length (m.alphabet)
  in
    for i = 0 to m.states - 1 do
      for j = 0 to m.states -1 do
	if i <> j (* All loops are at the end *)
	then
	  begin
	    let need_trans = ref false
	    in
	      for a = 0 to alpha - 1 do
		need_trans :=  !need_trans || (m.transitions.(a).(i).(j) <> False)
	      done;
	      if !need_trans
	      then
		print_string ("\\ArcL[.5]{State" ^
				(string_of_int i) ^
				"}{State" ^
				(string_of_int j) ^
				"}{" ^
				(transition_to_string m i j) ^
				"}\n")
	  end
      done
    done;
    (* Self-loops *)
    for i = 0  to m.states - 1 do
      let need_trans = ref false
      in
	for a = 0 to alpha - 1 do
	  need_trans := !need_trans || (m.transitions.(a).(i).(i) <> False)
	done;
	if !need_trans
	then
	  print_string ("\\LoopN[.5]{State" ^
				(string_of_int i) ^
				"}{" ^
				(transition_to_string m i i) ^
				"}\n")
    done;  
    print_string "\\end{VCPicture}}\n\\end{tabular}\n"
;;

(** printMtsTikz m prints the pgf/Tizk code for MTS m. Don't forget
\usepackage{tikz} and \usetikzlibrary{automata}.                    *)
let printMtsTikz (ismarkedmerge:bool) (m:mts) =
  (* Opening of the environment. *)
  print_string "\\begin{tikzpicture}[node distance=4cm,thick,initial text=,auto]\n\\tikzstyle{every state}=[draw=blue!50,very thick,fill=blue!20]\n";
  (* States *)
  let maxstatelength = String.length (string_of_int(m.states - 1))
  and maxprevstatelength = String.length (string_of_int(max 0 (m.states - 2)))
  and alpha = Array.length (m.alphabet)
  and maxact = ref 0
  in
    for a = 0 to (alpha - 1) do
      let newlength = String.length (m.alphabet.(a))
      in
	if newlength > !maxact then maxact := newlength
    done;
    for s = 0 to (m.states - 1) do
      let init = if s = 0 then ",initial] " else "]         "
      and statenb = string_of_int s
      and prevstatenb = string_of_int (s - 1)
      in
      let pad1 = padding (maxstatelength - (String.length statenb))
      and pad2 = 
	if s = 0 then
	  padding (18 + maxprevstatelength)
	else
	  padding (6 + maxprevstatelength - (String.length prevstatenb))
      in
      let position = if s = 0 then " " else (" [right of=q" ^ prevstatenb ^ "]")
      and statelabel =
	if ismarkedmerge then
	  begin
	    let (i,j) = correspStates m s
	    and mys2i x = if x < 0 then "*" else (string_of_int x)
	    in
	      ((mys2i i) ^ "," ^ (mys2i j) ^ "'")
	  end
	else
	  (string_of_int s)
      in
	print_string ("\n\\node[state" ^ init ^ "(q" ^ statenb ^ ")" ^ pad1 ^ position ^ pad2 ^ " {$" ^ statelabel ^ "$};");
    done;
  (* Transitions *)
    for i = 0 to (m.states - 1) do
      for j = 0 to (m.states - 1) do
	let orig = string_of_int i
	and dest = string_of_int j
	in
	let pad1 = padding (maxstatelength - (String.length orig))
	and mode =
	  if i = j then (* Loop *)
	    "[loop below]"
	  else
	    "[bend left] "
	and trans =
	  if ismarkedmerge then
	    unmarked_transition_to_string m i j
	  else
	    transition_to_string m i j
	in
	  if trans <> "" then
	    print_string ("\n\\path[->] (q" ^ orig ^ ")" ^ pad1 ^ " edge " ^ mode ^ " node {$" ^ trans ^ "$} (q" ^ dest ^ ");")
      done;
    done;
  (* Closing the environment *)
    print_string "\n\\end{tikzpicture}\n"    
;;

(** transition_to_fsp n m t returns a string representing a transition t of m in FSP.
States are ni for the i-th state, the initial state (0) being just the name n. The
transition is a tuple that contents the index of the action, the target, and the value
(True, Maybe, or False). The source is not important here: it is displayed elsewhere.  *)
let transition_to_fsp (name:string) (m:mts) ((letter,target,value):int*int*trieval) =
  let ponctuation = function
    | True  -> ""
    | Maybe -> "?"
    | False -> failwith "[transition_to_fsp] MTSs cannot display false transitions"
  in
    (String.uncapitalize m.alphabet.(letter)) ^ (ponctuation value) ^ " -> " ^ name ^ (if target <> 0 then string_of_int target else "")
;;

(** printMtsFsp n m prints the FSP code for the MTS m named n. We define one process
per state of the MTS, which is not the best way to define an MTS by hand. In order to
separate different states that don't have any outgoing transitions from one another
in MTSA, we add to them a self loop on an action called __epsilon__.                  *)
let printMtsFsp (raw_name:string) (m:mts) =
  let name = 
    if raw_name = ""
    then
      "MTS"
    else
      String.capitalize raw_name
  in
    print_string "\n/* FSP specification automatically generated: it IS NOT a pretty way to specify processes ! */\n";
    let max_digit = String.length (string_of_int (m.states - 1))
    in
      if m.states = 0
      then
	print_string (name ^ " = STOP.\n")
      else
	for i = 0 to m.states - 1 do
	  let pretty_out_trans = List.map (transition_to_fsp name m) (get_outgoing is_possible m i)
	  and the_pad = 
	    if i = 0
	    then
	      padding max_digit
	    else
	      padding (max_digit - String.length (string_of_int i))
	  and indent = "\n" ^ (padding ((String.length name) + max_digit + 3)) ^ "|" (* 3 = String.length " = " *)
	  and full_name = name ^ (if i=0 then "" else (string_of_int i))
	  in
	    print_string (full_name ^ the_pad ^ " = ");
	    if pretty_out_trans = []
	    then
	   (* print_string "END"                                  (* if we want all END spaces to be one *)*)
	      print_string ("(__epsilon__ -> " ^ full_name ^ ")") (* if we want to separate END states *)
	    else
	      print_string ("(" ^ (String.concat indent pretty_out_trans) ^ ")");
	    if i = m.states - 1
	    then
	      print_string ("\n\t+ " ^ (extractAlphaString "{" "}" "," m false) ^ ".\n")
	    else
	      print_string ",\n"
	done
;;

(** printMtsCaml m prints the code of the MTS as it is represented in this program.
Useful to keep parsed MTSs. *)
let printMtsCaml (m:mts) =
  let alphArray = extractAlphaString "[|\"" "\"|]" "\";\"" m false
  and alpha = Array.length m.alphabet
  and indent = padding 2
  in
  let dindent = indent ^ indent
  in
  let tindent = dindent ^ indent
  in
    print_string ("{\n" ^ indent ^ "alphabet = " ^ alphArray ^ ";\n" ^ indent ^ "states = " ^ (string_of_int m.states) ^ ";\n" ^ indent ^ "transitions = [|\n");
    for a = 0 to (alpha - 1) do
      print_string (indent ^ "(* On action '" ^ m.alphabet.(a) ^ "'. *)\n" ^ dindent ^ "[|");
      for i = 0 to (m.states -1) do
	print_string ((if i = 0 then "" else ("\n" ^ tindent)) ^ "[|");
	for j = 0 to (m.states - 1) do
	  print_string ((padded_of_trieval m.transitions.(a).(i).(j)) ^ (if j < (m.states - 1) then ";" else ""))
	done;
	print_string ("|]" ^ (if i < (m.states - 1) then ";" else ""));
      done;
      print_string "|]";
      if  (a < alpha - 1) then print_string ";\n\n"
    done;
    print_string ("\n" ^ indent ^ "|]\n}")
;;

(** printValue_matrix p m outputs the matrix as it is internally, with
labels as headers and with a precision of p digits after the point.    *)
let printValue_matrix (precision:int) (m:value_matrix) =
  if m <> [||]
  then
    let height = Array.length(m)
    and width = Array.length(m.(0))
    in
      (* We assume that values are between 0 and 1 *)
      let cell_length = max_list [(precision + 2);
				      (String.length (string_of_int (height-1)));
				      (String.length (string_of_int (width-1)))]
	(* For the '0' and the '.' *)
      in
      let hrule = build_hrule cell_length (width + 1)
      in
	print_string("\n"^hrule^"\n\t|" ^ (padding cell_length) ^ "|");
	for stateh = 0 to width-1 do
	  let to_print = string_of_int stateh
	  in
	    let bef,aft = split_space cell_length to_print
	    in
	      print_string((padding bef) ^ to_print ^ (padding aft) ^ "|")
	done;
	print_string("\n" ^ hrule ^ "\n");
	for statev = 0 to height-1 do
	  let to_print = string_of_int statev
	  in
	    let bef,aft = split_space cell_length to_print
	    in
	      print_string("\t|" ^ (padding bef) ^ to_print ^ (padding aft) ^ "|");
	      for stateh = 0 to width-1 do
		let cell_to_print = string_of_float (float_round  precision m.(statev).(stateh))
		in
		  let cell_bef,cell_aft = split_space cell_length cell_to_print
		  in
		    print_string((padding cell_bef) ^ cell_to_print ^ (padding cell_aft) ^ "|")
	      done;
	      print_string("\n" ^ hrule ^ "\n")
	done
;;

(** printValue_matrixLatex n p m outputs a LaTeX table representing
the matrix m whose values are rounded to p digits after the point.
The name n is displayed in the top left corner of the table. LaTeX
commands can be used for this, but the '\' need to be escaped.      *)
let printValue_matrixLatex (titre:string) (precision:int) (m:value_matrix) = 
 if m <> [||]
 then
   begin
     let height = Array.length(m)
     and width = Array.length(m.(0))
     and ruler = "\n\\\\ \\hline\n"
     in
       print_string "\n\\begin{tabular}{";
       for j = 0 to width do
	 print_string "|c"
       done;
       print_string ("|}\n\\hline\n"^titre);
       for j = 0 to width-1 do
	 print_string " & ";
	 print_int j;
       done;
       print_string ruler;
       for i = 0 to height-1 do
	 print_int i;
	 for j = 0 to width-1 do
	   print_string " & ";
	   print_float (float_round  precision m.(i).(j))
	 done;
	 print_string ruler
       done;
       print_string "\\end{tabular}\n"
   end
;;

(** printList f s l prints the contents of the list l using function f
separated by s.                                                        *)
let rec printList (subroutine:'a -> unit) (sep:string) = function
  |[] -> ()
  |t::[] -> subroutine t
  |t::q ->
    begin
     subroutine t;
     print_string sep;
     printList subroutine sep q
    end
;;

let printQueue (subroutine:'a -> unit) (sep:string) queue =
  let dup = Queue.copy queue
  in
    while not (Queue.is_empty dup) do
      let item = Queue.pop dup
      in
	subroutine item;
	print_string sep;
    done
;;

(** printIntPair (i,j) prints the string "(i,j)". *)
let printIntPair (i,j) =
  print_string ("(" ^ (string_of_int i) ^ "," ^ (string_of_int j) ^ ")")
;;

(** printPartialOrder po prints a partial order on pairs of states represented by a list
of lists of these pairs, top first. Incomparable pairs are printed on the same line.     *)
let printPartialOrder (po:(int * int) list list) =
  printList (printList printIntPair "\t") "\n" po
;;

let printPartialOrderLatex (po:(int * int) list list) =
  print_string "\\[\n\\begin{array}{c}\n";
  printList (printList printIntPair "\\quad") " \\\\\n" po;
  print_string "\\end{array}\n\\]\n"
;;

let rec printPath (m:mts) = function
  | [] -> ()
  | (o,l,d)::[] -> print_string ((string_of_int o) ^ " --" ^ m.alphabet.(l) ^ (if (is_maybe m.transitions.(l).(o).(d)) then "?" else "") ^ "--> " ^ (string_of_int d) ^ "\n")
  | (o,l,d)::q ->
      begin
	print_string ((string_of_int o) ^ " --" ^ m.alphabet.(l) ^ (if (is_maybe m.transitions.(l).(o).(d)) then "?" else "") ^ "--> ");
	printPath m q
      end
;;

let rec printPathMarked (m:mts) = function
  | [] -> ()
  | (o,l,d)::[] ->
      begin
	printIntPair (correspStates m o);
	print_string (" --" ^ m.alphabet.(l) ^ (if (is_maybe m.transitions.(l).(o).(d)) then "?" else "") ^ "--> ");
	printIntPair (correspStates m d);
	print_string "\n"
      end
  | (o,l,d)::q ->
      begin
	printIntPair (correspStates m o);
	print_string (" --" ^ m.alphabet.(l) ^ (if (is_maybe m.transitions.(l).(o).(d)) then "?" else "") ^ "--> ");
	printPathMarked m q
      end
;;

let rec printPathLatex (m:mts) = function
  | [] -> ()
  | (o,l,d)::[] -> print_string ((string_of_int o) ^ " \\" ^ (if (is_maybe m.transitions.(l).(o).(d)) then "may" else "true") ^ "transition{"^ m.alphabet.(l) ^ "} " ^ (string_of_int d) ^ "\n")
  | (o,l,d)::q ->
      begin
	print_string ((string_of_int o) ^ " \\" ^ (if (is_maybe m.transitions.(l).(o).(d)) then "may" else "true") ^ "transition{"^ m.alphabet.(l) ^ "} ");
	printPathLatex m q
      end
;;

let rec muCalcToString = function
  | Empty -> "$"
  | MuTrue -> "t"
  | MuFalse -> "f"
  | Var(x) -> x
  | Enext(a,f) -> "<"^a^">"^(muCalcToString f)
  | Anext(a,f)-> "["^a^"]"^(muCalcToString f)
  | Not(f)-> ""^(muCalcToString f)
  | Or(f1,f2) -> "("^(muCalcToString f1)^" v "^(muCalcToString f2)^")"
  | And(f1,f2) -> "("^(muCalcToString f1)^" ^ "^(muCalcToString f2)^")"
  | Mu(x,f) -> "mu "^x^"."^(muCalcToString f)
  | Nu(x,f) -> "nu "^x^"."^(muCalcToString f)
;;

let rec muCalcToLatex = function
  | Empty -> "\\bot"
  | MuTrue -> "\\mathbf{t}"
  | MuFalse -> "\\mathbf{f}"
  | Var(x) -> x
  | Enext(a,f) -> "\\senext{"^a^"} "^(muCalcToLatex f)
  | Anext(a,f)-> "\\sanext{"^a^"} "^(muCalcToLatex f)
  | Not(f)-> "\\neg "^(muCalcToLatex f)
  | Or(f1,f2) -> "("^(muCalcToLatex f1)^" \\vee "^(muCalcToLatex f2)^")"
  | And(f1,f2) -> "("^(muCalcToLatex f1)^" \\wedge "^(muCalcToLatex f2)^")"
  | Mu(x,f) -> "\\mu "^x^"."^(muCalcToLatex f)
  | Nu(x,f) -> "\\nu "^x^"."^(muCalcToLatex f)
;;

(* Getting good feedback: several tries. *)
let pathToScenario (name:string) (m:mts) (isMerge:bool) (path:(int * int * int) list) =
  let rec alphaString = extractAlphaString "{" "}" "," m isMerge
  and routine = function
    | [] -> ()
    | (o,l,d)::q ->
	begin
	  let a = if isMerge then (l - (l mod 3)) else l
	  in
	    print_string (m.alphabet.(a) ^ (if (is_maybe m.transitions.(l).(o).(d)) then "?" else "") ^ " -> ");
	    routine q
	end
  in
    print_string ("abstract "^ (String.capitalize name) ^ " = (");
    routine path;
    print_string ("END) + " ^ alphaString ^ ".\n")
;;

let pathToTree (name:string) (m:mts) (isMerge:bool) (path:(int * int * int) list) =
  let rec alphaString = extractAlphaString "{" "}" "," m isMerge
  and routine = function
    | [] -> 0
    | (o,l,d)::[] -> o
    | (o,l,d)::q ->
	begin
	  let a = if isMerge then (l - (l mod 3)) else l (* Remove guards. *)
	  in
	    print_string (m.alphabet.(a) ^ (if (is_maybe m.transitions.(l).(o).(d)) then "?" else "") ^ " -> ");
	    routine q
	end
  and printLastTrans = function
    | [] -> failwith "[pathToTree] There should be at least one transition out of the last state of the path."
    | t::[] -> print_string (m.alphabet.(t) ^ " -> ")
    | l -> 
	begin
	  print_string "{";
	  printList (fun a -> print_string m.alphabet.(a)) "," l;
	  print_string "} -> ";
	end
  and insert i = function
    | [] -> [i]
    | t::q when t = i -> t::q
    | t::q -> t::(insert i q)
  in
    print_string ("abstract " ^ (String.capitalize name) ^ " = (");
    let lastState = routine path
    in
    let outtrans = ref []
    and alpha = Array.length m.alphabet
    in
      for a = 0 to (alpha - 1) do
	let unguarded = if isMerge then (a - (a mod 3)) else a (* Remove guards. *)
	in
	  for j = 0 to (m.states - 1) do
	    if (is_true m.transitions.(a).(lastState).(j)) then outtrans := insert unguarded !outtrans;
	  done
      done;
      printLastTrans !outtrans;
      print_string ("END) + " ^ alphaString ^ ".\n")
;;

let pathToTreeBis (name:string) (m:mts) (isMerge:bool) (firstMark:string) (secondMark:string) (path:(int * int * int) list) =
  let rec alphaString = extractAlphaString "{" "}" "," m isMerge
  and routine = function
    | [] -> 0,0
    | (o,l,d)::[] -> o,l
    | (o,l,d)::q ->
	begin
	  let a = if isMerge then (l - (l mod 3)) else l (* Remove guards. *)
	  in
	    print_string (m.alphabet.(a) ^ (if (is_maybe m.transitions.(l).(o).(d)) then "?" else "") ^ " -> ");
	    routine q
	end
  and printLastTrans = function
    | [] -> failwith "[pathToTree] There should be at least one transition out of the last state of the path."
    | t::[] -> print_string (m.alphabet.(t) ^ " -> ")
    | l -> 
	begin
	  print_string "{";
	  printList (fun a -> print_string m.alphabet.(a)) "," l;
	  print_string "} -> ";
	end
  and insert i = function
    | [] -> [i]
    | t::q when t = i -> t::q
    | t::q -> t::(insert i q)
  in
    print_string ("abstract " ^ (String.capitalize name) ^ " = (");
    let lastState,label = routine path
    in
    let outtrans = ref []
    and alpha = Array.length m.alphabet
    and firstonly = ref false
    and secondonly = ref false
    in
      if isMerge then
	begin
	  let theAct = m.alphabet.(label)
	  and fstL = String.length firstMark
	  and sndL = String.length secondMark
	  in
	  let actL = String.length theAct
	  in
	    firstonly := ((String.sub theAct (actL - fstL) fstL) = firstMark);
	    secondonly := ((String.sub theAct (actL - sndL) sndL) = secondMark)
	end;
      for a = 0 to (alpha - 1) do
	let unguarded = if isMerge then (a - (a mod 3)) else a (* Remove guards. *)
	and doAllow = (not isMerge) || (!firstonly && (a mod 3 = 1)) || (!secondonly && (a mod 3 = 2))
	in
	  if doAllow then
	    for j = 0 to (m.states - 1) do
	      if (is_true m.transitions.(a).(lastState).(j)) then outtrans := insert unguarded !outtrans;
	    done
	  ;
      done;
      printLastTrans !outtrans;
      print_string ("END) + " ^ alphaString ^ ".\n")
;;

let pathToProperty (name:string) (m:mts) (isMerge:bool) (path:(int * int * int) list) =
  let rec alphaString = extractAlphaString "{" "}" "," m isMerge
  and routine first xs = function
    | [] -> ()
    | (o,l,d)::[] ->
	begin
	  let a = if isMerge then (l - (l mod 3)) else l
	  in
	    if not first then print_string " -> ";
	    print_string (xs ^ "!" ^ m.alphabet.(a))
	end
    | (o,l,d)::q ->
	begin
	  let a = if isMerge then (l - (l mod 3)) else l
	  in
	    if not first then print_string " && ";
	    print_string (xs ^ m.alphabet.(a));
	    routine false ("X " ^ xs) q
	end
  in
    print_string ("constraint " ^ (String.capitalize name) ^ " = (");
    routine true "" path;
    print_string (") + " ^ alphaString ^ "\n")
;;

let pathToMts (m:mts) (isMerge:bool) = function
  | [] -> failwith "[pathToMts] Empty path."
  | (_,l,_)::[] ->
      begin
	let alpha = if isMerge then (Array.length m.alphabet) / 3 else (Array.length m.alphabet)
	and noOLD a =
	  if (isMerge && (l / 3 = a)) || ((not isMerge) && (l = a))
	  then
	    [|[|False|]|]
	  else
	    [|[|Maybe|]|]
	in
	let res = {
	  alphabet = Array.init alpha (fun a -> if isMerge then m.alphabet.(a * 3) else m.alphabet.(a));
	  states = 1;
	  transitions = Array.init alpha noOLD
	}
	in
	  res
      end
  | list ->
      begin
	let rec lengthLast = function
	  | [] -> failwith "[pathToMts] Empty path."
	  | t::[] -> (0,t)
	  | t::q -> let prevlength,last = lengthLast q in (prevlength + 1,last)
	in
	let (leng,(lastState,label,_)) = (lengthLast list)
	and outtrans = ref []
	and alpha = Array.length m.alphabet
	in
	  for a = 0 to (alpha - 1) do
	    let unguarded = if isMerge then (a - (a mod 3)) else a (* Remove guards. *)
	    in
	      for j = 0 to (m.states - 1) do
		if (is_true m.transitions.(a).(lastState).(j)) then outtrans := insert unguarded !outtrans;
	      done
	  done;
	  let nbStates = leng + (List.length !outtrans) + 1
	  and alphaRes = if isMerge then (Array.length m.alphabet) / 3 else (Array.length m.alphabet)
	  in
	  let res = {
	    alphabet = Array.init alphaRes (fun a -> if isMerge then m.alphabet.(a * 3) else m.alphabet.(a));
	    states = nbStates;
	    transitions = Array.init alphaRes (fun a -> Array.init nbStates
						 (fun i -> Array.init nbStates
						    (fun j -> if j = (nbStates - 1) then Maybe else False)))
	  }
	  in
	  let rec buildPath n = function
	    | [] -> failwith "[pathToMts] Empty path."
	    | t::[] -> failwith "[pathToMts] Path of length 1 should have been matched before."
	    | (_,l,_)::t::[] ->
		begin
		  let unguarded = if isMerge then (l / 3) else l (* Remove guards. *)
		  in
		    res.transitions.(unguarded).(n).(nbStates - 1) <- False;
		    for j = leng to (nbStates - 2) do
		      res.transitions.(unguarded).(n).(j) <- Maybe;
		    done
		end
	    | (_,l,_)::q ->
		begin
		  let unguarded = if isMerge then (l / 3) else l (* Remove guards. *)
		  in
		    res.transitions.(unguarded).(n).(nbStates - 1) <- False;
		    res.transitions.(unguarded).(n).(n+1) <- Maybe;
		    buildPath (n + 1) q
		end
	  and forbidAct n = function
	    | [] -> ()
	    | t::q ->
		begin
		  let unguarded = if isMerge then (t / 3) else t (* Remove guards. *)
		  in
		    res.transitions.(unguarded).(n).(nbStates - 1) <- False;
		    forbidAct (n + 1) q
		end
	  in
	    buildPath 0 list;
	    forbidAct leng !outtrans;
	    res
      end
;;


let pathToMtsBis (m:mts) (isMerge:bool) (firstMark:string) (secondMark:string) = function
  | [] -> failwith "[pathToMts] Empty path."
  | (_,l,_)::[] ->
      begin
	let alpha = if isMerge then (Array.length m.alphabet) / 3 else (Array.length m.alphabet)
	and noOLD a =
	  if (isMerge && (l / 3 = a)) || ((not isMerge) && (l = a))
	  then
	    [|[|False|]|]
	  else
	    [|[|Maybe|]|]
	in
	let res = {
	  alphabet = Array.init alpha (fun a -> if isMerge then m.alphabet.(a * 3) else m.alphabet.(a));
	  states = 1;
	  transitions = Array.init alpha noOLD
	}
	in
	  res
      end
  | list ->
      begin
	let rec lengthLast = function
	  | [] -> failwith "[pathToMts] Empty path."
	  | t::[] -> (0,t)
	  | t::q -> let prevlength,last = lengthLast q in (prevlength + 1,last)
	in
	let (leng,(lastState,label,_)) = (lengthLast list)
	and outtrans = ref []
	and alpha = Array.length m.alphabet
	and firstonly = ref false
	and secondonly =ref false
	in
	  if isMerge then
	    begin
	      let theAct = m.alphabet.(label)
	      and fstL = String.length firstMark
	      and sndL = String.length secondMark
	      in
	      let actL = String.length theAct
	      in
		firstonly := ((String.sub theAct (actL - fstL) fstL) = firstMark);
		secondonly := ((String.sub theAct (actL - sndL) sndL) = secondMark)
	    end;
	  for a = 0 to (alpha - 1) do
	    let unguarded = if isMerge then (a - (a mod 3)) else a (* Remove guards. *)
	    and doAllow = (not isMerge) || (!firstonly && (a mod 3 = 1)) || (!secondonly && (a mod 3 = 2))
	    in
	      if doAllow then 
		for j = 0 to (m.states - 1) do
		  if (is_true m.transitions.(a).(lastState).(j)) then outtrans := insert unguarded !outtrans;
		done
	      ;
	  done;
	  let nbStates = leng + (List.length !outtrans) + 1
	  and alphaRes = if isMerge then (Array.length m.alphabet) / 3 else (Array.length m.alphabet)
	  in
	  let res = {
	    alphabet = Array.init alphaRes (fun a -> if isMerge then m.alphabet.(a * 3) else m.alphabet.(a));
	    states = nbStates;
	    transitions = Array.init alphaRes (fun a -> Array.init nbStates (fun i -> Array.init nbStates (fun j -> if j = (nbStates - 1) then Maybe else False)))
	  }
	  in
	  let rec buildPath n = function
	    | [] -> failwith "[pathToMts] Empty path."
	    | t::[] -> failwith "[pathToMts] Path of length 1 should have been matched before."
	    | (_,l,_)::t::[] ->
		begin
		  let unguarded = if isMerge then (l / 3) else l (* Remove guards. *)
		  in
		    res.transitions.(unguarded).(n).(nbStates - 1) <- False;
		    for j = leng to (nbStates - 2) do
		      res.transitions.(unguarded).(n).(j) <- Maybe;
		    done
		end
	    | (_,l,_)::q ->
		begin
		  let unguarded = if isMerge then (l / 3) else l (* Remove guards. *)
		  in
		    res.transitions.(unguarded).(n).(nbStates - 1) <- False;
		    res.transitions.(unguarded).(n).(n+1) <- Maybe;
		    buildPath (n + 1) q
		end
	  and forbidAct n = function
	    | [] -> ()
	    | t::q ->
		begin
		    res.transitions.(t).(n).(nbStates - 1) <- False;
		    forbidAct (n + 1) q
		end
	  in
	    buildPath 0 list;
	    forbidAct leng !outtrans;
	    res
      end
;;

(** pathToMu mpm upm m1 m2 p returns the mu-calculus (distingushing) formula
associated with the path p in the pseudomerge upm of m1 and m2. We also need
a marked version of the pseudo merge mpm to easily keep track of the path in
the original MTSs. The result is proven to be distinguishing in the case of
m1 and m2 being deterministic MTSs. The alphabets of m1 and m2 shoud be the
same (and in the same order), and the pseudo merge built with the function
pseudo_mergeMtsGen so that we can make assumptions over their alphabets.    *)
let rec pathToMu (markedmerge:mts) (unmarkedmerge:mts) (mts1:mts) (mts2:mts) = function
  | [] -> Empty
  | (_,l,_)::[] ->
      begin
	let from = l mod 3
	in
	let letter =  unmarkedmerge.alphabet.(l - from)
	in
	  match from with
	    | 0 -> failwith "[pathToMu] Path does not end in disagreement"
	    | 1 -> Enext(letter,MuTrue)
	    | _ -> Anext(letter,MuFalse)
      end
  | (o,l,d)::q ->
      begin
	let tail = pathToMu markedmerge unmarkedmerge mts1 mts2 q
	and from = l mod 3
	and backindex = correspStates markedmerge
	in
	let letter = if from = 0 then l else failwith "[pathToMu] Guarded transition along the path"
	in
	let letterorig = letter / 3 (* It is important that mts1 and mts2  have the same alphabet in the same order. *)
	and realletter = unmarkedmerge.alphabet.(letter)
	and o1,o2 =  backindex o
	and d1,d2 = backindex d
	in
	  if is_true mts1.transitions.(letterorig).(o1).(d1)
	  then
	    Enext(realletter,tail)
	  else if is_true mts2.transitions.(letterorig).(o2).(d2)
	  then 
	    Anext(realletter,tail)
	  else
	    failwith "[pathToMu] Path has maybe transitions: it is NOT an inconsistency witness."
      end
;;

(** pathToMu pm m1 m2 p returns the mu-calculus (distingushing) formula
associated with the path p in the pseudomerge pm of m1 and m2. The result
is proven to be distinguishing in the case of m1 and m2 being deterministic
MTSs. The alphabets of m1 and m2 shoud be the same (and in the same order),
and the pseudo merge built with the function pseudo_mergeMtsGen so that we
can make assumptions over its alphabet. Cleaner than pathToMu.           *)
let pathToMuNomark (merge:mts) (mts1:mts) (mts2:mts) (path:(int * int * int) list) =
  let rec routine current1 current2 = function
    | [] -> Empty
    | (_,l,_)::[] ->
      begin
	let from = l mod 3
	in
	let realletter =  merge.alphabet.(l - from)
	in
	  match from with
	    | 0 -> failwith "[pathToMu] Path does not end in disagreement"
	    | 1 -> Enext(realletter,MuTrue)
	    | _ -> Anext(realletter,MuFalse)
      end
    | (o,l,d)::q ->
	begin
	  let from = l mod 3
	  in
	  let letter = if from = 0 then l else failwith "[pathToMuNomark] Disagreement before the end of the path."
	  in
	  let letterorig = letter  / 3
	  and realletter = merge.alphabet.(letter)
	  and out1 = ref []
	  and out2 = ref []
	  in
	    (* Looking for matching outgoing transitions in the original models. *)
	    for i = 0 to mts1.states - 1 do
	      let trans = mts1.transitions.(letterorig).(current1).(i)
	      in
		if is_possible trans
		then out1 := (i,trans)::(!out1)
	    done;
	    for i = 0 to mts2.states - 1 do
	      let trans = mts2.transitions.(letterorig).(current2).(i)
	      in
		if is_possible trans
		then out2 := (i,trans)::(!out2)
	    done;
	    let newcurrent1,status1 =
	      match !out1 with
		| [] -> failwith "[pathToMuNomark] Disagreement before the end of the path. Moreover, this does not appear in the pseudo-merge."
		| t::[] -> t
		| _ -> failwith "[pathToMuNomark] First model is not deterministic."
	    and newcurrent2,status2 =
	      match !out2 with
		| [] -> failwith "[pathToMuNomark] Disagreement before the end of the path. Moreover, this does not appear in the pseudo-merge."
		| t::[] -> t
		| _ -> failwith "[pathToMuNomark] Second model is not deterministic."
	    in
	    let tail = routine newcurrent1 newcurrent2 q
	    in
	      if status1 = True then
		Enext(realletter,tail)
	      else if status2 = True then
		Anext(realletter,tail)
	      else
		failwith "[pathToMuNomark] Path has maybe transitions: it is NOT an inconsistency witness."
	end
  in
    routine 0 0 path
;;

(** Adaptation of the distinguish algorithm so that the formula goes along a path. *)
let pathToMuNonDet (marked:mts) (merge:mts) (mts1:mts) (mts2:mts) (simvals:value_matrix) path =
  let debug = false in
  let m = mts1.states
  and n = mts2.states
  in
  let placeholders = Array.init m (fun i -> Array.make n None)
  and pushed = Array.init m (fun i -> Array.make n false)
  and visited = Array.init m (fun i -> Array.make n false)
  and queue = Queue.create()
  and debug1 a i = if debug then print_string("Accessing "^a^".("^(string_of_int i)^")\n")
  and debug2 a i j = if debug then print_string("Accessing "^a^".("^(string_of_int i)^").("^(string_of_int j)^")\n")
  and debug3 a i j k= if debug then print_string("Accessing "^a^".("^(string_of_int i)^").("^(string_of_int j)^").("^(string_of_int k)^")\n")
  in
  let rec translatePath = function
    | [] -> []
    | (o,l,d)::q ->
	begin
	  let backindex = correspStates marked
	  in
	    (backindex o,l/3,backindex d)::(translatePath q)
	end
  and searchPath (s1,s2) = function
    | [] -> (-1,-1),-1,(-1,-1)
    | ((t1,t2),l,d)::q when t1 = s1 && t2=s2 -> (t1,t2),l,d
    | t::q -> searchPath (s1,s2) q
  and union l = function
    | [] -> l
    | t::q -> union (insert t l) q
  and ancestorsMerge queue i j f =
    let tmpqueue = Queue.create()
    in
      while (not (Queue.is_empty queue)) do
	let k,l,m = Queue.pop queue
	in
	  if (k = i && l = j) then
	      Queue.push (k,l,(union m f)) tmpqueue
      done;
      Queue.transfer tmpqueue queue
  and cleverpush (i,j,f) thequeue =
    debug2 "pushed" i j;
    if not pushed.(i).(j) then
      begin
	Queue.push (i,j,f) thequeue;
	pushed.(i).(j) <- true
      end
    else
      ancestorsMerge thequeue i j f
  and getOutlistWithLetter mts letter state =
    let list_out = ref []
    in
      for j = 0 to (mts.states - 1) do
	debug3 "mts.transitions" letter state j;
	if is_possible mts.transitions.(letter).(state).(j)
	then list_out := j::(!list_out)
      done;
      !list_out
  and allNew f = function
    | [] -> true
    | (s1,s2,f0)::q ->
	(not (List.mem (s1,s2) f)) && (allNew f q)
  and inconsistentList first s = function
    | [] -> true
    | t::q ->
	begin
	  if first then
	    (simvals.(s).(t) < 1.) && (inconsistentList first s q)
	  else
	    (simvals.(t).(s) < 1.) && (inconsistentList first s q)
	end
  and toInconsist f (outr1,outp2) (outr2,outp1) =
    if debug then List.iter printIntPair f;
    let rec candidates isok pos = function
      |[] -> []
      | (l,t,_)::q -> 
	  begin
	    let cand1 = candidatesone (l,t) pos
	    in
	      if isok t cand1 then
		(l,t)::(candidates isok pos q)
	      else
		candidates isok pos q
	  end
    and candidatesone (l1,t1) = function
      | [] -> []
      | (l2,t2,_)::q when l1 = l2 -> t2::(candidatesone (l1,t1) q)
      | t::q -> candidatesone (l1,t1) q
    and ok1 f t1 = function
      |[] -> true
      | t2::q ->
	  begin
	    (*debug2 "visited" t1 t2;
	    if debug then print_string ("Value: "^(string_of_bool (visited.(t1).(t2)))^"\n");*)
	    debug2 "simvals" t1 t2;
	    if debug then print_string ("Value: "^(string_of_float (simvals.(t1).(t2)))^"\n");
	    (not (List.mem (t1,t2) f)) && (simvals.(t1).(t2) < 1.) && (ok1 f t1 q)
	  end
    and ok2 f t1 = function
      |[] -> true
      | t2::q ->
	  begin
	    (*debug2 "visited" t2 t1;
	    if debug then print_string ("Value: "^(string_of_bool (visited.(t2).(t1)))^"\n");*)
	    debug2 "simvals" t2 t1;
	    if debug then print_string ("Value: "^(string_of_float (simvals.(t2).(t1)))^"\n");
	    (not (List.mem (t2,t1) f)) && (simvals.(t2).(t1) < 1.) && (ok2 f t1 q)
	  end
    in
    let candidates1 = candidates (ok1 f) outp2 outr1
    and candidates2 = candidates (ok2 f) outp1 outr2
    in
      match candidates1 with
	| [] ->
	    begin
	      match candidates2 with
		| [] -> failwith "[pathToMuNonDet][toInconsist] No inconsistency found."
		| (l,t2)::q -> false,l,t2
	    end
	| (l,t1)::q -> true,l,t1
	    
  and formulaFromPh (orig1,orig2) =
    (*debug2 "[formulaFromPh] placehoders" orig1 orig2;*)
    match placeholders.(orig1).(orig2) with
      |None -> Empty
      |Some(l,t,child) ->
	 begin
	   if t then
	     Enext(l,buildChildCunj child)
	   else
	     Anext(l,buildChildDisj child)
	 end
  and buildChildCunj = function
    | [] -> MuTrue
    | pair::[] -> formulaFromPh pair
    | pair::q -> And(formulaFromPh pair,buildChildCunj q)
  and buildChildDisj = function
    | [] -> MuFalse
    | pair::[] -> formulaFromPh pair
    | pair::q -> Or(formulaFromPh pair,buildChildDisj q)
  in
  let translatedPath = translatePath path
  in
    cleverpush (0,0,[(0,0)]) queue;
    while not (Queue.is_empty queue) do
      let c1,c2,f = Queue.pop queue
      and notOk = ref true
      in
      let (o1,o2),l,(d1,d2) = searchPath (c1,c2) translatedPath
      in
	debug2 "visited" c1 c2;
	visited.(c1).(c2) <- true;
	if ((o1 = c1) && (o2 = c2))
	then
	  begin
	    if debug then print_string "(* Case when we want to follow a specific letter. *)\n";
	    debug3 "mts1.transitions" l o1 d1;
	    if (d1 >= 0) && is_true mts1.transitions.(l).(o1).(d1) then
	      (* Case when the required transition is in the 1st MTS. *)
	      begin
		let outs = getOutlistWithLetter mts2 l o2
		in
		let destph = List.map (fun x -> d1,x,(insert (d1,x) f)) outs
		in
		  if debug then printList (fun (x,y,z) -> printIntPair (x,y)) ";" destph;
		  if (allNew f destph) && (inconsistentList true d1 outs) then
		    begin
		      notOk := false;
		      List.iter (fun triple -> cleverpush triple queue) destph;
		      debug2 "placeholders" o1 o2;
		      debug1 "mts1.alphabet" l;
		      placeholders.(o1).(o2) <- Some (mts1.alphabet.(l),true,(List.map (fun (x,y,z) -> x,y) destph))
		    end;
	      end;
	    if (!notOk && (debug3 "mts2.transitions" l o2 d2; (d2 >= 0)) && is_true mts2.transitions.(l).(o2).(d2)) then
	      (* Case when the required transition is in the 2nd MTS. *)
	      begin
		let outs = getOutlistWithLetter mts1 l o1
		in
		let destph = List.map (fun x -> x,d2,(insert (x,d2) f)) outs
		in
		  if debug then printList (fun (x,y,z) -> printIntPair (x,y)) ";" destph;
		  if (allNew f destph) && (inconsistentList false d2 outs) then
		    begin
		      notOk := false;
		      List.iter (fun triple -> cleverpush triple queue) destph;
		      debug2 "placeholders" o1 o2;
		      debug1 "mts2.alphabet" l;
		      placeholders.(o1).(o2) <- Some (mts2.alphabet.(l),false,(List.map (fun (x,y,z) -> x,y) destph))
		    end;
	      end;
	    (*
	    if not (!notOk) then
	      ignore (Queue.pop pathtogo);
	    else
	      failwith ("[pathToMuNonDet] No required transition found on letter "^(mts1.alphabet.(l))^" at state pair ("^(string_of_int o1)^","^(string_of_int o2)^").")
	    *)
	  end;
	if !notOk then
	  begin
	    if debug then print_string ("(* Case whe we use the normal distinguish algorithm for pair ("^(string_of_int c1)^","^(string_of_int c2)^"). *)\n");
	    let out1r = get_outgoing is_true mts1 c1
	    and out2r = get_outgoing is_true mts2 c2
	    and out1p = get_outgoing is_possible mts1 c1
	    and out2p = get_outgoing is_possible mts2 c2
	    in
	    let first,letter,state = toInconsist f (out1r,out2p) (out2r,out1p)
	    in
	      if debug then print_string ("Result of toInconsist: ("^(string_of_bool first)^","^(string_of_int letter)^","^(string_of_int state)^").\n");
	      if first then
		begin
		  let outs = getOutlistWithLetter mts2 letter c2
		  in
		  let destph = List.map (fun x -> state,x,(insert (state,x) f)) outs
		  in
		    List.iter (fun triple -> cleverpush triple queue) destph;
		    debug2 "placeholders" c1 c2;
		    debug1 "mts1.alphabet" letter;
		    placeholders.(c1).(c2) <- Some (mts1.alphabet.(letter),true,(List.map (fun (x,y,z) -> x,y) destph))
		end
	      else
		begin
		  let outs = getOutlistWithLetter mts1 letter c1
		  in
		  let destph = List.map (fun x -> x,state,(insert (x,state) f)) outs
		  in
		    List.iter (fun triple -> cleverpush triple queue) destph;
		    debug2 "placeholders" c1 c2;
		    debug1 "mts2.alphabet" letter;
		    placeholders.(c1).(c2) <- Some (mts2.alphabet.(letter),false,(List.map (fun (x,y,z) -> x,y) destph))
		end
		
	  end
    done;
    formulaFromPh (0,0)
;;



(* Checking the input *)

(** checkMts m returns 'true' iff m is a well formed MTS. We check that
the letters do not contain a question mark or a coma, which would bring
confusion, and that they are unique. We also check that the size of the
transitions tables match the length of the alphabet and the number of states. *)
let checkMts (m:mts) =
  let alpha = Array.length(m.alphabet)
  in
    let ok = ref (alpha = Array.length (m.transitions))
    and a = ref 0
    in
      while (!ok && !a < alpha) do
	(* We proscribe the use of '?' and ',' in the actions *)
	ok := !ok &&
	  (not (String.contains m.alphabet.(!a) '?')) &&
	  (not (String.contains m.alphabet.(!a) ','));
	(* For each letter we check that we have a m.states-sized matrix *)
	let letter_matrix = m.transitions.(!a)
	in
	  let height = Array.length(letter_matrix)
	  and k = ref 0
	  in
	    ok := !ok && (height = m.states);
	    while (!ok && !k < height) do
	      ok := !ok && (Array.length(letter_matrix.(!k)) = m.states);
	      k := !k + 1
	    done;
	    (* And that there is no letter doublon *)
	    let b = ref 0
	    in
	      while (!ok && !b < !a) do
		ok := !ok && (m.alphabet.(!a) <> m.alphabet.(!b));
		b :=  !b + 1
	      done;
	      a := !a + 1
      done;
      !ok
;;


(** checkMatrix m returns 'true' iff m is a matrix (all lines have the same length). *)
let checkMatrix (m:'a array array) =
  let height = Array.length(m)
  in
    if (height = 0)
    then
      true
    else
      let width = Array.length(m.(0))
      and good = ref true
      and i = ref 1
      in
	while (!good && !i<height) do
	  good := (Array.length(m.(!i)) = width);
	  i := !i + 1
	done;
	!good
;;

(** checkActSet m n returns 'true' iff the alphabets of the inputs
contain the same letters. It assumes the inputs are correct MTSs.  *)
let checkActSet (mtsM:mts) (mtsN:mts) =
  let vocM = mtsM.alphabet
  and vocN = mtsN.alphabet
  in
    let alphaM = Array.length(vocM)
    and alphaN = Array.length(vocN)
    in
      if alphaM <> alphaN
      then
	false
      else
	let i = ref 0
	and ok = ref true
	in
	  while (!ok && !i < alphaM) do
	    let goOn = ref true
	    and j = ref 0
	    in
	      while (!goOn && !j < alphaN) do
 		if vocM.(!i) = vocN.(!j)
		then
		  goOn := false;
		j := !j + 1
	      done;
	      ok := not !goOn;
	      i := !i + 1
	  done;
	  !ok
;;

(** reorderAct m n if M and N have the same alphabet, returns a fresh MTS which is semantically
speaking identical to N (in terms of MTS) but has the same order of letters as M. Fails otherwise. *)
let reorderAct (mtsM:mts) (mtsN:mts) =
  let vocM = mtsM.alphabet
  and vocN = mtsN.alphabet
  in
    let alphaM = Array.length(vocM)
    and alphaN = Array.length(vocN)
    in
      if alphaM <> alphaN
      then
	failwith "[reorderAct] Different size alphabets"
      else
	let corresp = Array.make alphaM (-1)
	and i = ref 0
	and ok = ref true
	in
	  while (!ok && !i < alphaM) do
	    let goOn = ref true
	    and j = ref 0
	    in
	      (* Building the permutation (corresp), if possible *)
	      while (!goOn && !j < alphaN) do
 		if vocM.(!i) = vocN.(!j)
		then
		  begin
		    corresp.(!i) <- !j;
		    goOn := false
		  end;
		j := !j + 1
	      done;
	      ok := not !goOn;
	      i := !i + 1
	  done;
	  if not !ok
	  then
	    failwith "[reorderAct] Different alphabets"
	  else
	    begin
	      (* Building a copy of mtsN with reordered transitions_matrix according to the permutation *)
	      let newN = initMts mtsM.alphabet mtsN.states
	      in
		for a = 0 to alphaM-1 do
		  for i = 0 to mtsN.states-1 do
		    for j = 0 to mtsN.states-1 do
		      newN.transitions.(a).(i).(j) <- mtsN.transitions.(corresp.(a)).(i).(j);
		    done
		  done;
		done;
		newN
	    end
;;

(** checkAct m n returns 'true' iff the alphabets of the input
MTSs are identical, not only in content, but also in order.    *)
let checkAct (mtsM:mts) (mtsN:mts) =
  let vocM = mtsM.alphabet
  and vocN = mtsN.alphabet
  in
    let alphaM = Array.length(vocM)
    and alphaN = Array.length(vocN)
    and ok = ref true
    and k = ref 0
    in
      ok := alphaM = alphaN;
      while (!ok && !k < alphaM) do
	ok := !ok && (vocM.(!k) = vocN.(!k));
	k := !k + 1
      done;
      !ok
;;

(** checkValidity m n v returns 'true' iff the input MTSs m and n are well formed,
over the *exact* same alphabet, and the matrix p corresponds to a similarity  matrix
for the MTSs (we check that sizes match. We do not accept an empty MTSs or matrix.   *)
let checkValidity (mtsM:mts) (mtsN:mts) (matr:value_matrix) =
  let ok = ref ((checkMts mtsM) && (checkMts mtsN) && (checkAct mtsM mtsN) && (checkMatrix matr))
  and dimM = mtsM.states
  and dimN = mtsN.states
  and height = Array.length(matr)
  in
    ok := !ok && (dimM > 0) && (dimN > 0) && (height > 0);
    if !ok
    then
      begin
	let width = Array.length(matr.(0))
	in
	  ok := !ok && (dimM = height) && (dimN = width);
	  let i = ref 0
	  in
	    while (!ok && !i < height) do
	      let j = ref 0
	      in
		while (!ok && !j < width) do
		  ok := !ok && (matr.(!i).(!j) >= 0.) && (matr.(!i).(!j) <= 1.);
		  j := !j + 1
		done;
		i := !i + 1
	    done;
      end;
    !ok
;;

let checkPath (m:mts) (path:(int * int * int) list) =
  let rec checkPathRec prev = function
    | [] -> true
    | (o,l,d)::q -> (prev = o) && (is_possible m.transitions.(l).(o).(d)) && (checkPathRec d q)
  in
    checkPathRec 0 path
;;

(* Various tools part II (use checking functions) *)

(** toMts r uses the raw input parsed by FspParser and returns
a fresh MTS corresponding to this data, along with its name.   *)
let toMts (fromParse:parseResult) =
 (*  let name = String.uncapitalize fromParse.name
  let name = fromParse.name in *)
  let nbstates = fromParse.nbstates
  and rawTrans = fromParse.trans
  in
  let rec insert i = function
    | [] -> [i]
    | t::q when t = i -> t::q
    | t::q -> t::(insert i q)
  in
  let rec extractAlpha already = function
    | [] -> already
    | (_,outlist)::q -> extractAlpha (subExtractAlpha already outlist) q
  and subExtractAlpha already = function
    | [] -> already
    | (acts,_)::q -> subExtractAlpha (insertAllActs already acts) q
  and insertAllActs already = function
    | [] -> already
    | (_,t)::q -> insertAllActs (insert t already) q
  in
  let alphabet = Array.of_list (extractAlpha [] rawTrans)
  in
  let alpha = Array.length alphabet
  in
  let alphaIndex act =
    let curr = ref 0
    in
      while ((!curr < alpha) && (alphabet.(!curr) <> act)) do
	curr :=  !curr + 1
      done;
      if (!curr = alpha)
      then
	failwith "Action not found in the alphabet"
      else
	!curr
  in
  let transitions = Array.init alpha (fun x -> Array.init nbstates (fun y -> Array.make nbstates False))
  in
  let rec extractTrans = function
    | [] -> ()
    | (origin,outlist)::q ->
	begin
	  extractOut origin outlist;
	  extractTrans q
	end
  and extractOut orig = function
    | [] -> ()
    | (acts,dest)::q ->
	begin
	  extractBetween orig dest acts;
	  extractOut orig q
	end
  and extractBetween orig dest = function
    | [] -> ()
    | (poss,act)::q ->
	begin
	  transitions.(alphaIndex act).(orig).(dest) <- poss;
	  extractBetween orig dest q
	end
  in
    extractTrans rawTrans;
    let result = {
      alphabet = alphabet;
      states = nbstates;
      transitions = transitions}
    in
      if checkMts result
      then
	fromParse.name,result
      else
	failwith "Input did not yield a correct MTS."
;;


(** transpose m returns a fresh matrix which is the transposition of m. *)
let transpose (matr:value_matrix) =
  if not (checkMatrix matr)
  then
    failwith "[transpose] Matrix not a matrix"
  else
    if matr = [||]
    then
      ([||]:value_matrix)
    else
      let height = Array.length(matr)
      and width = Array.length(matr.(0))
      in
        let result = Array.make_matrix width height 0.
	in
	  for i = 0 to height-1 do
	    for j = 0 to width-1 do
	      result.(j).(i) <- matr.(i).(j)
	    done
	  done;
	  (result:value_matrix)
;;

(** extractConsistency s t m  takes a value_matrix m and a threshold t and returns
all pairs of integers which corresponding value in m is over the threshold. It can
be specified if the threshold is to be strict (s is 'true') or not (s is 'false'). *)
let extractConsistency (strict:bool) (threshold:float) (m:value_matrix) =
  if not ((checkMatrix m) && (threshold >= 0.) && (threshold <= 1.))
  then
    failwith "[extractConsistency] Invalid input"
  else
    begin
      let res = ref []
      and height = Array.length m
      in
	if height > 0
	then
	  begin
	    let width = Array.length m.(0)
	    in
	      for i = 1 to height do
		for j = 1 to width do
		  let value = m.(height-i).(width-j)
		  in
		    (*print_string ("Extracting pseudo consistency (threshold = " ^ (string_of_float threshold) ^ "): (" ^ (string_of_int (height-i)) ^ "," ^ (string_of_int (width-j)) ^ "): " ^ (string_of_float value) ^ " -> "); *)
		    if (strict && (value > threshold)) || ((not strict) && (value >= threshold))
		    then
                      (*begin*)
			res := (height-i,width-j)::(!res);
			(*print_string "kept.\n"
		      end
		    else
		      print_string "discarded.\n"*)
		done
	      done;
	  end;
	!res
    end
;;

(** abstractMts m s t l returns a fresh MTS that abstracts the input one m as follows:
 - if the target t is a known state, the transition s--l--> t is turned to Maybe if it
was True or False (fails otherwise).
 - if t is an unknown state (index = number of states in the input), we add it before
doing the previous.                                                                    *)
let abstractMts (refined:mts) (source:int) (target:int) (label:int) =
  if ((not (checkMts refined)) || (source < 0 ) || (source >= refined.states) || (target < 0) || (target > refined.states) || (label < 0) || (label >= Array.length refined.alphabet))
  then
    failwith "[abstractMts] Invalid input MTS or abstraction command"
  else
    begin
      if target = refined.states
      then
	begin
	  (* We must add a * state *)
	  let trans_copy a i j =
	    if (i < refined.states) && (j < refined.states)
	    then
	      refined.transitions.(a).(i).(j)
	    else
	      False
	  and alpha = Array.length refined.alphabet
	  in
	  let res = {
	    alphabet = Array.copy refined.alphabet;
	    states = refined.states + 1;
	    transitions = Array.init alpha (fun a -> Array.init (refined.states + 1) (fun i -> Array.init (refined.states + 1) (trans_copy a i)))
	  }
	  in
	    res.transitions.(label).(source).(target) <- Maybe;
	    (*(* Maybe self-loops on all letter for the new state (We after all don't want to) *)
	    for a = 0 to alpha - 1 do
	      res.transitions.(a).(refined.states).(refined.states) <- Maybe
	    done;*)
	    res
	end
      else
	begin
	  let res = copyMts refined
	  in
	    begin
	      match refined.transitions.(label).(source).(target) with
		| Maybe -> failwith "[abstractMts] Invalid abstraction: this transition is already 'Maybe'"
		| _ -> res.transitions.(label).(source).(target) <- Maybe
	    end;
	    res
	end;
    end
;;

(** pseudo_mergeMtsGen s1 s2 m1 m2 c1 c2 takes two MTSs m1 and m2, a 'pseudo-consistency'
relation c1, and a real consistency relation c2.
The merge will yield some transitions coming from inconsistencies: they will be guarded by
marking the transitions with s1 (resp. s2) if the transition is only in m1 (resp. m2). The
merge will also create some states that are product of a real and a special state.
Unreachable states are removed. This operator is supposed to be used for visualisation of
a merge, not as an actual one. The pair (0,0) should always be in c: otherwise we fail.       *)
let pseudo_mergeMtsGen (keepTrack:bool) (firstmark:string) (secondmark:string) (mts1:mts) (mts2:mts) (consistency:(int * int) list) (real_consistency:(int * int) list) =
  let rec mtsOk = (checkMts mts1) && (checkMts mts2) && (checkAct mts1 mts2)
  and consistencyOk = function
    | [] -> true
    | (i,j)::q -> (i >= 0) && (i < mts1.states) && (j >= 0) && (j < mts2.states) && (consistencyOk q)
  and in_list e = function
    | [] -> false
    | t::q when t=e -> true
    | t::q -> in_list e q
  in
    if not (mtsOk && (consistencyOk consistency))
    then
      failwith "[pseudo_mergeMtsGen] Incorrect input"
    else if not (in_list (0,0) consistency)
    then
      failwith "[pseudo_mergeMtsGen] The models are too inconsistent: (0,0) is not in the pseudo-consistency relation"
    else
      begin
	let rec alpha = Array.length mts1.alphabet
	and new_letter k =
	  let res = ref mts1.alphabet.(k/3)
	    (* The marks should be distinct: otherwise we don't know in which MTS is a guarder transition *)
	  and realfirstmark = if firstmark = secondmark then (firstmark ^ "1") else firstmark
	  and realsecondmark = if firstmark = secondmark then (secondmark ^ "2") else secondmark
	  in
	    begin
	      match k mod 3 with
		| 1 -> res := !res ^ realfirstmark
		| 2 -> res := !res ^ realsecondmark
		| _ -> ()
	    end;
	    !res
	in
	  (* We have an alphabet three times bigger as the initial one:
	  we have the actions that are guarded in one MTS or another.   *)
	let new_alphabet_act = Array.init (3 * alpha) (new_letter)
	in
	let new_alphabet = ref new_alphabet_act
	  (* We cannot know by advance the number of states that are going to be
	  reachable. But it is always under (mts1.states + 1) * (mts2.states + 1).
	  state_space: the set S in the algorithm.                                 *)
	and nb_new_states = ref 0
	and statespace = Array.init (mts1.states + 1) (fun x -> Array.make (mts2.states + 1) (-1))
	  (* to_do_list: the set P in the algorithm. *)
	and to_do_list = ref []
	  (* true_transitions [union]  may_transitions: the set T in the algorithm. *)
	and true_transitions = ref []
	and may_transitions = ref []
	in
	  (* add_trans: adds a transition on a from (s1,s2) to (t1,t2) in the merge. If (t1,t2)
	  is new, it is added to the statespace and the to_do_list. The transition is 'possible':
	  it is a true one iff possible = True, and a maybe one iff possible = Maybe. It should
	  never be called to add a false transition, and in that case will do nothing.            *)
	let add_trans a s1 s2 t1 t2 possible =
	  if possible <> False
	  then
	    begin
	      if statespace.(t1).(t2) = -1
	      then
		begin
		  statespace.(t1).(t2) <- !nb_new_states;
		  nb_new_states := !nb_new_states + 1;
		  to_do_list := !to_do_list@[t1,t2]
		end;
	      if possible = True
	      then
		true_transitions := ((s1,s2),a,(t1,t2))::(!true_transitions)
	      else (* possible = Maybe *)
		may_transitions := ((s1,s2),a,(t1,t2))::(!may_transitions)
	    end
	in
	  (* We add the initial pair of states into the merge and the to_do_list *)
	  statespace.(0).(0) <- !nb_new_states;
	  nb_new_states := !nb_new_states + 1;
	  to_do_list := (0,0)::(!to_do_list);
	  (* The main loop: we traverse all reachable pair of states *)
	  while !to_do_list <> [] do
	    match !to_do_list with
	      | (s1,s2)::q -> (* This case will always happen due to the test when entering the loop *)
		  begin
		    to_do_list := q; (* 'pop' *)
		    if (s1 = mts1.states) && (s2 = mts2.states)
		    then
		      failwith "[pseudo_mergeMtsGen] State (*,*) reached. (Should not happen)"
		    else if s1 = mts1.states (* we are only in the second MTS *)
		    then
		      begin
			for a = 0 to alpha -1 do
			  for t2 = 0 to mts2.states - 1 do
			    let out =  mts2.transitions.(a).(s2).(t2)
			    in
			      if out <> False (* Or just when out = True ? (Depends if we
					      want to display the guarded maybe transitions). *)
			      then
				add_trans a mts1.states s2 mts1.states t2 out
			  done
			done
		      end
		    else if s2 = mts2.states (* we are only in the first MTS *)
		    then
		      begin
			for a = 0 to alpha -1 do
			  for t1 = 0 to mts1.states - 1 do
			    let out =  mts1.transitions.(a).(s1).(t1)
			    in
			      if out <> False (* Or just when out = True ? (Depends if we
					      want to display the guarded maybe transitions). *)
			      then
				add_trans a s1 mts2.states t1 mts2.states out
			  done
			done
		      end
		    else (* We are still in a 'normal' pair of states *)
		      begin
			for a = 0 to alpha - 1 do
			  let outgoing1 = ref []
			  and outgoing2 = ref []
			  in
			    for t1 = 0 to mts1.states - 1 do
			      if mts1.transitions.(a).(s1).(t1) <> False
			      then
				outgoing1 := t1::(!outgoing1)
			    done;
			    for t2 = 0 to mts2.states - 1 do
			      if mts2.transitions.(a).(s2).(t2) <> False
			      then
				outgoing2 := t2::(!outgoing2)
			    done;
			    (* compare_out b l1 l2: for every transition in l2 checks if there
			    is at least a matching transition in l1. If there are, creates the
			    corresponding ones in the merge. Otherwise creates a guarded one.
			    b is only there to identify which list corresponds to which MTS.   *)
			    let rec compare_out first_mts out = function
			      | [] -> ()
			      | t::q ->
				  begin
				    if not (find_all_suitable first_mts t out)
				    then (* We have not found any suitable matching transition *)
				      if first_mts
				      then
					add_trans a s1 s2 t mts2.states True
				      else
					add_trans a s1 s2 mts1.states t True;
				    compare_out first_mts out q
				  end
			    (* find_all_suitable does the check for compare_out for one transition of l2.
			    It adds all suitable transitions, and returns true iff there is no problem.   *)
			    and find_all_suitable first_mts t = function
			      | [] -> (* No problem if the transition was Maybe (here we won't have False transitions) *)
				  if first_mts
				  then
				    mts1.transitions.(a).(s1).(t) = Maybe
				  else
				    mts2.transitions.(a).(s2).(t) = Maybe
			      | u::q ->
				  begin
				    (* rest will add all other suitable transitions *)
				    let rest = find_all_suitable first_mts t q
				    in
				      if first_mts
				      then
					if in_list (t,u) consistency
					then (* We have a matching transition *)
					  begin
					    if (mts1.transitions.(a).(s1).(t) = True) || (mts2.transitions.(a).(s2).(u) = True)
					    then
					      add_trans a s1 s2 t u True
					    else
					      if in_list (t,u) real_consistency then add_trans a s1 s2 t u Maybe;
					    true
					  end
					else
					  rest
				      else
					if in_list (u,t) consistency
					then (* We have a matching transition *)
					  begin
					    if (mts1.transitions.(a).(s1).(u) = True) || (mts2.transitions.(a).(s2).(t) = True)
					    then
					      add_trans a s1 s2 u t True
					    else
					      if in_list (t,u) real_consistency then add_trans a s1 s2 u t Maybe;
					    true
					  end
					else
					  rest
				  end
			    in
			      (* Since we have to do twice very analogous things, we just
			      treated the differences with the 'first_mts' boolean flag.  *)
			      compare_out true !outgoing2 !outgoing1;
			      compare_out false !outgoing1 !outgoing2
			done
		      end
		  end
	      | _ -> () (* Will never happen because of the condition when entering the 'while' loop *)
	  done;
	  (* Now we can put the pseudo-merge in the right form (of type mts) *)
	  (* If we want to keep track of were the states came from in the merge
	     with self loops on the states, we will need a bigger alphabet.     *)
	  let alphaSize = ref 0
	  in
	    if keepTrack
	    then
	      alphaSize := 3*alpha + !nb_new_states
	    else
	      alphaSize := 3*alpha
	    ;
	  let new_transitions = Array.init (!alphaSize) (fun y -> Array.init !nb_new_states (fun y -> Array.make !nb_new_states False))
	  in
	  (* add_trans_in_new will add the transitions in a list in the corresponding
	     transitions_matrix, with the value 'possible' (True or Maybe in this case). *)
	  let rec add_trans_in_new possible = function
	    | [] -> ()
	    | ((s1,s2),a,(t1,t2))::q ->
		let actual_letter = 
		  if t1 = mts1.states (* we are only in the second MTS *)
		  then
		    3*a+2
		  else if t2 = mts2.states (* we are only in the first MTS *)
		  then
		    3*a+1
		  else (* we don't have a guarded transition *)
		    3*a
		and s = statespace.(s1).(s2)
		and t = statespace.(t1).(t2)
		in
		  if (s = -1) || (t = -1)
		  then
		    failwith "[pseudo_mergeMtsGen] We are reaching an unknown state. (Should not happen)"
		  else
		    new_transitions.(actual_letter).(s).(t) <- possible;
		  add_trans_in_new possible q
	  in
	    (* We actually transfer the transitions from the lists to the matrices. *)
	    add_trans_in_new Maybe !may_transitions;
	    add_trans_in_new True !true_transitions;
	    (* We add the labels as self loops. We also need to modify the alphabt accordingly. *)
	    if keepTrack
	    then
	      begin
		let copy_new_alphabet_act a =
		  if ((a >= 0) && (a < 3*alpha))
		  then new_alphabet_act.(a)
		  else ""
		and state_to_string limit s =
		  if (s >= limit)
		  then "new"
		  else (string_of_int s)
		in
		new_alphabet := Array.init (!alphaSize) copy_new_alphabet_act;
		for i = 0 to mts1.states do
		  for j = 0 to mts2.states do
		    let new_state = statespace.(i).(j)
		    in
		      if new_state >= 0
		      then
			begin
			  new_transitions.(3*alpha+new_state).(new_state).(new_state) <- True;
			  (!new_alphabet).(3*alpha+new_state) <- ("__" ^ (state_to_string mts1.states i) ^ "_" ^ (state_to_string mts2.states j) ^ "__");
			end
		  done
		done
	      end;
	    (* We finally build the merge as an mts object. *)
	    let merge = {
	      alphabet = !new_alphabet;
	      states = !nb_new_states;
	      transitions = new_transitions
	    }
	    in
	      merge
      end
;;

let pseudo_mergeMts =
  pseudo_mergeMtsGen false
;;

let extract_disagreements (pm:mts) =
  let alpha = Array.length pm.alphabet
  in
  let rec agreeDisagree agree disagree = function
    | -1 -> (agree,disagree)
    | n ->
	if isStar pm n then
	  agreeDisagree agree (n::disagree) (n-1)
	else
	  agreeDisagree (n::agree) disagree (n-1)
  and disagreebetween agState disState = function
    | -1 -> []
    | a ->
	if is_possible pm.transitions.(a).(agState).(disState) then
	  (agState,pm.alphabet.(a),disState)::(disagreebetween agState disState (a-1))
	else
	  disagreebetween agState disState (a-1)
  and disagreefromag agState = function
    | [] -> []
    | disState::q -> (disagreebetween agState disState (alpha-1))@(disagreefromag agState q)
  and disagree allDis = function
    | [] -> []
    | agState::q -> (disagreefromag agState allDis)@(disagree allDis q)
  in
  let ag,dis = agreeDisagree [] [] (pm.states - 1)
  in
    disagree dis ag
;;
	

(* Computing similarities -- The heuristics *)

(** half_formula m n l v s t computes the $X_p^{(i+1)}(s,t)$, s
being a state of the MTS m and t being one of the MTS n. v is
a matrix that contains the values of the $Q_p^{(i)}(s',t')$. The
label similarity function l can be used if we were to assume that
some transitions can be different but still be similar. In the
examples we une only one such function: label_sim.                *)
let  half_formula (mtsM:mts) (mtsN:mts) (label_sim:label -> label -> float) (previous_values:value_matrix) (s:int) (t:int) =
  if not (checkValidity mtsM mtsN previous_values)
  then
    failwith "[half_formula] Invalid input"
  else
    begin
      let height = mtsM.states
      and width = mtsN.states
      and voc = mtsM.alphabet
      in
	if (s >= height) || (s < 0) || (t >= width) || (t < 0)
	then
	  failwith "[half_formula] Invalid states"
	else
	  begin
	    (* res: the final result. *)
	    let res = ref 0.
	    (* true_out: the n of the formula. *)
	    and true_out = ref 0
	    (* counted_out: n+|S_p(s,t)| in the formula. *)
	    and counted_out = ref 0
	    and alpha = Array.length voc
	    in
	      (* We first calculate the sum of Q_p(s',t') for the true transitions. *)
	      for a = 0 to alpha-1 do
		for i = 0 to height-1 do
		  let the_Mtransition = mtsM.transitions.(a).(s).(i)
		  in
		    if (the_Mtransition = True)
		    then
		      begin
			true_out := !true_out + 1;
			counted_out := !counted_out + 1;
			let lb_list = ref []
			and labelM = (the_Mtransition,voc.(a))
			in
			  (* We go through all possible transitions in mtsN. *)
			  for b = 0 to alpha-1 do
			    for j = 0 to width-1 do
			      let the_Ntransition = mtsN.transitions.(b).(t).(j)
			      in
				if (the_Ntransition <> False)
				then
				  begin
				    let labelN = (the_Ntransition,voc.(b))
				    in
				      let lb = (label_sim labelM labelN) *. previous_values.(i).(j)
				      in
					(* We build the set of Q_p(s',t'). *)
					lb_list := lb::(!lb_list)
				  end
			    done
			  done;
			  if !lb_list <> []
			  then
			    (* We add the max of the set of the Q_(s',t'). *)
			    res := !res +. (max_list !lb_list)
		      end
		done
	      done;
	      let onlytrue_rate = ref 1.
	      in
		if !true_out <> 0
		then
		  (* onlytrue_rate is the average of the consistency brought by required transitions. *)
		  onlytrue_rate := !res /. (float_of_int !true_out);
		(* We now go through the maybe transitions of mtsM: it is a lot similar to what
		we do with required transitions, except we don't always count an outgoing
		transition nor do we always add the maximal value of the Q_p(s',t') to the result. *)
		for a = 0 to alpha-1 do
		  for i = 0 to height-1 do
		    let the_Mtransition = mtsM.transitions.(a).(s).(i)
		    in
		      if (the_Mtransition = Maybe)
		      then
			begin
			  let lb_list = ref []
			  and labelM = (the_Mtransition,voc.(a))
			  in
			    for b = 0 to alpha-1 do
			      for j = 0 to width-1 do
				let the_Ntransition = mtsN.transitions.(b).(t).(j)
				in
				  if (the_Ntransition <> False)
				  then
				    begin
				      let labelN = (the_Ntransition,voc.(b))
				      in
				        let lb = (label_sim labelM labelN) *. previous_values.(i).(j)
					in
					  lb_list := lb::(!lb_list)
				    end
			      done
			    done;
			    if !lb_list <> []
			    then
			      begin
				let the_max = max_list !lb_list
				in
				  (* We only consider the transition if we have an
				  improvement compared to only the required transitions. *)
				  if (the_max >= !onlytrue_rate)
				  then
				    begin
				      counted_out := !counted_out + 1;
				      res := !res +. the_max
				    end
			      end
			end
		  done
		done;
	      !res,!counted_out
	  end
    end
;;

(** formula p m n l v s t computes the $Q_p^{(i+1)}(s,t)$, s
being a state of the MTS m and t being one of the MTS n. v is
a matrix that contains the values of the $Q_p^{(i)}(s',t')$. The
label similarity function l can be used if we were to assume that
some transitions can be different but still be similar. In the
examples we une only one such function: label_sim.                *)
let formulaAdd (weight:float) (mtsM:mts) (mtsN:mts) (label_sim:label -> label -> float) (previous_values:value_matrix) (s:int) (t:int) =
  if not (checkValidity mtsM mtsN previous_values)
  then
    failwith "[formulaAdd] Invalid input"
  else
    begin
      (* bigX = X_p(s,t); outM = n + |S_p(s,t)|; bigY = X_p(t,s); outN = m + |S_p(t,s)|. *)
      let bigX,outM = half_formula mtsM mtsN label_sim previous_values s t
      and bigY,outN = half_formula mtsN mtsM (fun x -> fun y -> label_sim y x) (transpose previous_values) t s
      in
	if (outM + outN = 0)
	then (* No outgoing transition from s and t. *)
	  previous_values.(s).(t)
	else (* We apply the formula for the case when we have some successors. *)
	  let succ = float_of_int (outM + outN)
	  in
	    weight *. ((bigX +. bigY) /. succ) +. (1. -. weight)
    end
;;

let formulaAve (weight:float) (mtsM:mts) (mtsN:mts) (label_sim:label -> label -> float) (previous_values:value_matrix) (s:int) (t:int) =
  if not (checkValidity mtsM mtsN previous_values)
  then
    failwith "[formulaAve] Invalid input"
  else
    begin
      (* bigX = X_p(s,t); outM = n + |S_p(s,t)|; bigY = X_p(t,s); outN = m + |S_p(t,s)|. *)
      let bigX,outM = half_formula mtsM mtsN label_sim previous_values s t
      and bigY,outN = half_formula mtsN mtsM (fun x -> fun y -> label_sim y x) (transpose previous_values) t s
      and quotM = ref 1.
      and quotN = ref 1.
      in
	if (outM > 0) then quotM := bigX /. (float_of_int outM);
	if (outN > 0) then quotN := bigY /. (float_of_int outN);
	weight *. (!quotM +. !quotN) /. 2. +. (1. -. weight)
    end
;;

let formulaMin (weight:float) (mtsM:mts) (mtsN:mts) (label_sim:label -> label -> float) (previous_values:value_matrix) (s:int) (t:int) =
  if not (checkValidity mtsM mtsN previous_values)
  then
    failwith "[formulaMin] Invalid input"
  else
    begin
      (* bigX = X_p(s,t); outM = n + |S_p(s,t)|; bigY = X_p(t,s); outN = m + |S_p(t,s)|. *)
      let bigX,outM = half_formula mtsM mtsN label_sim previous_values s t
      and bigY,outN = half_formula mtsN mtsM (fun x -> fun y -> label_sim y x) (transpose previous_values) t s
      and quotM = ref 1.
      and quotN = ref 1.
      and min = ref 1.
      in
	if (outM > 0) then quotM := bigX /. (float_of_int outM);
	if (outN > 0) then quotN := bigY /. (float_of_int outN);
	if !quotM > !quotN
	then
	  min := !quotN
	else
	  min := !quotM
	;
	weight *. !min +. (1. -. weight)
    end
;;

(** normalise p m takes a parameter p and a matrix of Q_p(s,t) and
returns a fresh matrix whose values are the real similarity value
(between 0 and 1) given by $Sim(s,t) = \frac{Q_p(s,t)-p}{1-p}$.    *)
let normalise (p:float) (m:value_matrix) =
  let height = Array.length m
  in
    if height = 0
    then
      [||]
    else
      let width = Array.length m.(0)
      in
	Array.init height (fun i -> Array.init width (fun j -> (m.(i).(j) +. p -. 1.)/.p))
;;

(** The similarity value between two labels that is being used:
different letters are totally different (0), and same letters are similar
iff they agree: True and False disagree, otherwise there is agreement.    *)
let label_sim (labelM:label) (labelN:label) =
  let valM,actM = labelM
  and valN,actN = labelN
  in
    if actM <> actN
    then
      0.
    else
      match valM,valN with
	| True ,True  -> 1.
	| True ,Maybe -> 1.
	| True ,False -> 0.
	| Maybe,True  -> 1.
	| Maybe,Maybe -> 1.
	| Maybe,False -> 1.
	| False,True  -> 0.
	| False,Maybe -> 1.
	| False,False -> 1.
;;


(* Iteration of a formula *)
(**iterate f m n l h i iterates the formula f for Q_p (we actually only use one)
i times for MTSs m and n. l is the label similarity (we don't use several ones),
and we could force some pairs of state to have a fixed Q_p by using the list h
(actually not used in the final heuristics). Returns the Q_p, not the normalized
matrix. The parameter p is actually embedded in the definition of the formula f. *)
let iterate formula (mtsM:mts) (mtsN0:mts) (label_sim:label -> label -> float) (forced_values:(int * float * int) list) nb_iter =
  if nb_iter < 0
  then
    failwith "[iterate] Negative iteration"
  else
    if not ((checkMts mtsM) && (checkMts mtsN0))
    then
      failwith "[iterate] Invalid MTS input"
    else
      begin
	let mtsN_temp = ref mtsN0
	in
	  if not (checkAct mtsM mtsN0)
	  then
	    mtsN_temp := reorderAct mtsM mtsN0; (* Will raise an exception if the alphabets do not correspond. *)
	  let mtsN = !mtsN_temp
	  in
	  let height = mtsM.states
	  and width = mtsN.states
	  in
          let current = Array.make_matrix height width 1.
	  and previous = Array.make_matrix height width 0.
	  and forced = Array.make_matrix height width false
	  and temp_list = ref forced_values
	  in
	    while !temp_list <> [] do
	      match !temp_list with
		|(s,v,t)::q ->
		   begin
		     if (s >= height) || (s < 0) || (t >= width) || (t < 0) || (v > 1.) || (v < 0.)
		     then
		       failwith "[iterate] Invalid forced values"
		     else
		       begin
			 forced.(s).(t) <- true;
			 current.(s).(t) <- v
		       end;
		     temp_list := q
		   end
		|[] -> ()
	    done;
	    for k = 1 to nb_iter do
	      (* Copy the current into the previous *)
	      for i = 0 to height-1 do
		for j = 0 to width -1 do
		  previous.(i).(j) <- current.(i).(j)
		done
	      done;
	      (* Compute the new values *)
	      for i = 0 to height-1 do
		for j = 0 to width-1 do
		  if not forced.(i).(j)
		  then
		    let new_value = formula mtsM mtsN label_sim previous i j
		    in
		      current.(i).(j) <- new_value
		done
	      done;
	    done;
	    current
      end
;;

(** iterate_lazy r f m n l h i works the same as iterate, but we stop since
we reached a fixpoint up to a precision of 10^{-r} (to be sure we iterate
again once). The computation is stopped after i iterations no matter what.  *)
let iterate_lazy (precision:int) formula (mtsM:mts) (mtsN0:mts) (label_sim:label -> label -> float) (forced_values:(int * float * int) list) max_iter =
  if ((max_iter < 0) || (precision <=0))
  then
    failwith "[iterate_lazy] Negative iteration"
  else
    if not ((checkMts mtsM) && (checkMts mtsN0))
    then
      failwith "[iterate_lazy] Invalid MTS input"
    else
      begin
	let mtsN_temp = ref mtsN0
	in
	  if not (checkAct mtsM mtsN0)
	  then
	    mtsN_temp := reorderAct mtsM mtsN0; (* Will raise an exception if the alphabets do not correspond. *)
	  let mtsN = !mtsN_temp
	  in
	  let height = mtsM.states
	  and width = mtsN.states
	  in
          let current = Array.make_matrix height width 1.
	  and previous = Array.make_matrix height width 0.
	  and forced = Array.make_matrix height width false
	  and temp_list = ref forced_values
	  and nb_iter = ref 0
	  and go_on = ref true
	  and go_two = ref true
	  in
	    while !temp_list <> [] do
	      match !temp_list with
		|(s,v,t)::q ->
		   begin
		     if (s >= height) || (s < 0) || (t >= width) || (t < 0) || (v > 1.) || (v < 0.)
		     then
		       failwith "[iterate_lazy] Invalid forced values"
		     else
		       begin
			 forced.(s).(t) <- true;
			 current.(s).(t) <- v
		       end;
		     temp_list := q
		   end
		|[] -> ()
	    done;
	    while (!nb_iter < max_iter && !go_on) do
	      go_on := !go_two;
	      nb_iter := !nb_iter + 1;
	      (* Copy the current into the previous *)
	      for i = 0 to height-1 do
		for j = 0 to width -1 do
		  previous.(i).(j) <- current.(i).(j)
		done
	      done;
	      (* Compute the new values *)
	      for i = 0 to height-1 do
		for j = 0 to width-1 do
		  if not forced.(i).(j)
		  then
		    let new_value = formula mtsM mtsN label_sim previous i j
		    in
		      current.(i).(j) <- new_value
		done
	      done;
	      (* Check for a value that changed above the certain precision *)
	      go_two := false;
	      for i = 0 to height-1 do
		for j = 0 to width-1 do
		  go_two := !go_two || ((float_round precision previous.(i).(j)) <> (float_round precision current.(i).(j)))
		done
	      done
	    done;
	    if !nb_iter = max_iter then print_string ("Warning : [iterate_lazy] Maximal number of iteration (" ^ (string_of_int max_iter) ^ ") was reached before the desired precision (10^-" ^ (string_of_int precision) ^ ") could be reached. The result might not be accurate. You may want to let more iterations or lower the precision.");
	    current
      end
;;



(* Testing the impact of parameters in the formula *)

(** extractPartialOrder m returns a partial ordering on the pairs (i,j) of coordinates
in m as a list of lists of such pairs where the inner lists represent pairs with an equal
value and the order in the outter list is corresponds to the order between the equivalence
classes, higher first. Inside an inner list, pairs are ordered lexicographically.          *)
let extractPartialOrder (m:value_matrix) =
  if (not (checkMatrix m))
  then
    failwith "[extractPartialOrder] Invalid input"
  else
    begin
      let rec length = Array.length(m)
      and width = Array.length(m.(0))
      and result = ref []
      and insert (i,j) = function
        |[] -> [[(i,j)]]
        |t::q ->
          begin
            match t with
              |[] -> failwith "[extractPartialOrder] Empty class"
              |(a,b)::r when m.(a).(b) > m.(i).(j) -> t::(insert (i,j) q)
              |(a,b)::r when m.(a).(b) < m.(i).(j) -> [(i,j)]::t::q
              |(a,b)::r -> ((i,j)::t)::q
          end
      in
        for i = length-1 downto 0 do
          for j = width-1 downto 0 do
            result := insert (i,j) !result
          done
        done;
        !result
    end
;;

let orderConsistency formula (cut:int) (precision:int) (p:float) (mts1:mts) (mts2:mts) =
  let result = normalise p (iterate_lazy precision (formula p) mts1 mts2 label_sim [] cut)
  in
    extractPartialOrder result
;;

let boundThresholdPrecomputed (result:value_matrix) (mts1:mts) (mts2:mts) =
  let looseConsistency = extractConsistency false 0. result
  in
  let looseMerge = pseudo_mergeMtsGen true "[1]" "[2]" mts1 mts2 looseConsistency []

  and insert i l =
    let rec insertRec = function
      | [] -> [i]
      | t::q when t > i -> t::(insertRec q)
      | t::q when t < i -> i::t::q
      | t::q -> t::q
    in
      if i <= result.(0).(0)
      then
	insertRec l
      else
	l
  in
  let reachable = ref []
  and thresholds = ref []
  and alpha = (Array.length looseMerge.alphabet)
  in
  let labelsStart = alpha - looseMerge.states
  in
    for a = labelsStart to alpha-1 do
      let i,j = parseLabel looseMerge.alphabet.(a)
      in
	if ((i >= 0) && (j >= 0)) then
	  begin
	    reachable := (i,j)::(!reachable);
	    thresholds := insert result.(i).(j) (!thresholds);
	  end
    done;
    !reachable,!thresholds
;;

let boundThreshold formula (cut:int) (precision:int) (p:float) (mts1:mts) (mts2:mts) =
  let result = normalise p (iterate_lazy precision (formula p) mts1 mts2 label_sim [] cut)
  in
    boundThresholdPrecomputed result mts1 mts2;;


(* Getting "useful" feedback. *)

let extractPath (m:mts) (isMerge:bool) (orig:int) (dest:int) (labelString:string) =
  (* Checking the input *)
  if ((orig >= m.states) || (dest >= m.states)) then failwith "[extractPath] Unknown state.";
  let findIndex (a:'a array) (goal:'a) =
    let n = Array.length a
    and k = ref 0
    in
      while (((!k) < n) && (a.(!k) <> goal)) do
	k := !k + 1
      done;
      if (!k = n)
      then failwith "[extractPath][findIndex] Not found."
      else !k
  in
  let rec (*insert (l,d) = function
      | [] -> [(l,d)]
      | (tl,td)::q when tl > l -> (l,d)::(tl,td)::q
      | (tl,td)::q when tl < l -> (tl,td)::(insert (l,d) q)
      | (tl,td)::q -> (tl,td)::q
  and *)isPresent e = function
    | [] -> false
    | t::q when t = e -> true
    | t::q -> isPresent e q
  in
  let successors o =
    let res = ref []
    and alpha = Array.length m.alphabet
    in
      for l = 0 to (alpha - 1) do
	if ((not isMerge) || (isMerge && (l mod 3 = 0)))
	then
	  for d = 0 to (m.states - 1) do
	    if is_true m.transitions.(l).(o).(d) (* Should we only allow True paths ? *)
	    then res := insert (l,d) !res
	  done
      done;
      !res      
  in
  let label = findIndex m.alphabet labelString
  and queue = Queue.create()
  and visited = ref [0]
  and found = ref (orig = 0)
  and res = ref []
  in
    if not (is_possible m.transitions.(label).(orig).(dest)) then failwith "[extractPath] Goal transition does not exists.";
    Queue.push (0,[]) queue;
    while not ((!found) || (Queue.is_empty queue)) do
      let localOrig,path = Queue.pop queue
      in
	(* print_string ((string_of_int localOrig) ^ "\t");*)
      let rec succ = successors localOrig
      and processSucc = function
	| [] -> ()
	| (l,d)::q when (not !found) ->
	    begin
	      if (not (isPresent d !visited))
	      then
		begin
		  found := !found || (d = orig);
		  let newPath = (localOrig,l,d)::path
		  in
		    if d = orig then res := newPath;
		    Queue.push (d,newPath) queue;
		    visited := d::(!visited);
		end;
	      processSucc q
	    end
	| _ -> ()
      in
(*
	printList (fun (x,y) -> print_string ((string_of_int localOrig) ^ " --" ^ (m.alphabet.(x)) ^ "--> " ^ (string_of_int y))) "\n" succ;
	print_string "\n";
*)
	processSucc succ;
(*
	print_string "queue: ";
	Queue.iter (fun (x,y) ->
		      print_string ("(" ^ (string_of_int x) ^ ",[");
		      printList (fun (a,b,c) ->
				   print_string ((string_of_int a) ^ " --" ^ (m.alphabet.(b)) ^ "--> " ^ (string_of_int c))
				) " | " y;
		      print_string "]) #") queue;
	print_string "\nvisited: ";
	printList print_int " | " !visited;
	print_string "\nsuccessors: ";
	printList (fun (x,y) -> print_int y) " | " succ;
	print_string "\n"
*)
    done;
    if (not !found) then failwith "[extractPath] No path found.";
    res := (orig,label,dest)::(!res);
    let rec mirror acc = function
      | [] -> acc
      | t::q -> mirror (t::acc) q
    in
      mirror [] !res
;;


(* Macros for an easier use of the algorithm *)

let computeSimVal (formula:float ->  mts ->  mts -> (label -> label -> float) -> value_matrix -> int -> int -> float) (cut:int) (precision:int) (p:float) (mts1:mts) (mts2:mts) =
  normalise p (iterate_lazy precision (formula p) mts1 mts2 label_sim [] cut)
;;

(** testParameters f c k p t b m n outputs the normalized similarity
values for m and n (for LaTeX iff b is 'true') for a precision of k
digits after the point and for a parameter p of p, using the formula f.
Returns the pseudo-consistency relation that contains all pair of
state with a similarity > t.                                            *)
let testParameters (formula:float ->  mts ->  mts -> (label -> label -> float) -> value_matrix -> int -> int -> float) (cut:int) (precision:int) (p:float) (threshold:float) (latex:bool) (mts1:mts) (mts2:mts) =
  if ((p <= 0.) || (p >= 1.) || (threshold < 0.) || (threshold > 1.))
  then
    failwith "[testParameters] Incorrect values: 0 < p < 1 and 0 <= threshold <= 1"
  else
    begin
      let result = normalise p (iterate_lazy precision (formula p) mts1 mts2 label_sim [] cut)
      in
        if latex
        then
          printValue_matrixLatex "" precision result
        else
          printValue_matrix precision result;
        extractConsistency false threshold result
    end
;;

(** testFormula b m n outputs the normalized similarity values for m
and n (for LaTeX iff b is 'true') for a precision of 5 digits after
the point and for a parameter p of 1/2. Returns the pseudo-consistency
relation that contains all pair of state with a similarity > 0.5.       *)
let testFormula  =
  testParameters formulaAdd 512 5 0.5 0.5
;;

(** testPseudoMerge o s m n c outputs an MTS which is the pseudo-merge of m and n
according to the consistency relation returned by testFormula, adding into it the
pair of states in c. The form of the output is determined by o: if it is True, the
output is just a table of states and transitions between them (see printMts). If
it is Maybe, the output will be LaTeX code (see printMtsLatex). If o is False, the
output will be FSP specification (see PrintMtsFsp).                                *)
let testPseudoMerge (output:trieval) (name:string) (mts1:mts) (mts2:mts) (additional_consistency:(int * int) list) =
  let consistency = additional_consistency@(testFormula false mts1 mts2)
  in
    match output with
      | True  -> (* For direct viewing (as a table of transitions *)
	  printMts (pseudo_mergeMts "[1]" "[2]" mts1 mts2 consistency [])
      | Maybe -> (* For LaTeX (using Vaucanson *)
	  printMtsLatex 3. name (pseudo_mergeMts "[1]" "[2]" mts1 mts2 consistency [])
      | False -> (* For MTSA (FSP) *)
	  printMtsFsp name (pseudo_mergeMts "_1_" "_2_" mts1 mts2 consistency [])
;;


let benchmarkParameter formula (beamer:bool) (cut:int) (precision:int) (interval:float) (mts1:mts) (mts2:mts) =
  let printAsFigure p =
    print_string "\n\\begin{figure}\n";
    printPartialOrderLatex (orderConsistency formula cut precision p mts1 mts2);
    print_string ("\\caption{Partial order of consistency between pairs with \\(p = " ^ (string_of_float p) ^ "\\)}\n\\end{figure}\n")
  and printAsFrame p =
    print_string ("\n\\begin{frame}\n\\frametitle{Partial order of consistency between pairs with \\(p = " ^ (string_of_float p) ^ "\\)}\n");
    printPartialOrderLatex (orderConsistency formula cut precision p mts1 mts2);
    print_string "\\end{frame}\n";
  in
    let currP = ref interval
    in
      while ((!currP > 0.) && (!currP < 1.)) do
	if beamer
	then printAsFrame !currP
	else printAsFigure !currP
	;
        currP := (!currP) +. interval;
      done;
;;


let benchmarkThreshold formula (cut:int) (precision:int) (p:float) (step:float) (name:string) (mts1:mts) (mts2:mts) =
  let heuristics = (normalise p (iterate_lazy precision (formula p) mts1 mts2 label_sim [] cut))
  in
  print_string "\n/*\n  Values computed by the heuristics :";
  printValue_matrix 5 heuristics;
  print_string "*/\n";
  let printAsFsp thresh =
    print_string ("\n// --- Pseudo merge for " ^ name ^ " with threshold = " ^ (string_of_float thresh) ^ " ---");
    let consistency = extractConsistency false thresh heuristics
    and threshString = string_of_float thresh
    in
    let escapedName = name ^ "_0point" ^ (String.sub threshString 2 ((String.length threshString) - 2)) ^ "_"
    in
      (*printList printIntPair " | " consistency;*)
      printMtsFsp escapedName (pseudo_mergeMtsGen true "_1" "_2" mts1 mts2 consistency []);
      print_string "// --- *** ---\n"
  and currThresh = ref 0.
  in
    while !currThresh < heuristics.(0).(0) do
      printAsFsp !currThresh;
      currThresh := (!currThresh) +. step
    done
;;
